-- TODO: rename Buff to Aura

TRAININGGROUNDS_Buff = {}
TRAININGGROUNDS_Buff.__index = TRAININGGROUNDS_Buff

function TRAININGGROUNDS_Buff:new()
	local self={}
	setmetatable(self, TRAININGGROUNDS_Buff)
		
	return self
end

function TRAININGGROUNDS_Buff:setup(combat,executiontime)
	--print("TRAININGGROUNDS_Buff:setup")

	self.combat=combat
	
	self:setCustomBuffInfo(combat)
	
	self.dying=false
	self.dead=false
	
	self.starttime=executiontime
	if(self.duration~=nil)then
		self.expirytime=executiontime+self.duration	
	end
		
	self.x=0
	self.y=0
	
	self.nextticktime=nil
	self.channeltickfrequency=nil
	self.ticksremaining=nil
	
	local existingbuff=nil
	
	--print("Buff:setup checking for existing "..tostring(self:class()))
	existingbuff=combat:GetBuffIfExists(self:class())
	
	-- for i=1,#self.combat.buffs,1 do		
		-- if(self.combat.buffs[i]:isa(self:class()))then
			-- existingbuff=self.combat.buffs[i]			
			-- break
		-- end
	-- end
	
	--TODO: ask buff if it extends the old buff or replaces it
		-- for example, RJW would replace it
	if(existingbuff~=nil)then
		--print("already exists")
		self:OnReapply(existingbuff,executiontime)
	else
		tinsert(self.combat.newbuffs,self)
	end
	
	if(self.visibletoplayer==true) then
		self.frame=TRAININGGROUNDS_ReusableFrames.getFrame()	
		self.frame:SetParent(self.combat.character.environment.frame)
		self.frame.framelevel=self.combat.character.environment.frame.framelevel+1	
		self.frame:SetFrameLevel(self.frame.framelevel)			
		self.frame:SetWidth(32);self.frame:SetHeight(32)
		self.frame.texture:SetTexture(self.icon)	
		self.frame.texture:Show()
		
		self.stacksframe=TRAININGGROUNDS_ReusableFrames.getFrame()	
		self.stacksframe:SetParent(self.frame)
		self.stacksframe.framelevel=self.frame.framelevel+1
		self.stacksframe:SetFrameLevel(self.stacksframe.framelevel)		
		self.stacksframe:SetWidth(32);self.stacksframe:SetHeight(32)
	end
	
	
	
end

function TRAININGGROUNDS_Buff:setCustomBuffInfo()
	-- override
	self.duration=1000
	self.visibletoplayer=true
	self.icon="Interface\\Icons\\inv_misc_questionmark"
	
	self.tickonbuffstart=false
	self.tickfrequency=nil
	--self.totalticks=nil
	
	self.durationaffectedbyhaste=false
	self.tickfrequencyaffectedbyhaste=false
	--TODO: partial ticks are NOT implemented yet

end

function TRAININGGROUNDS_Buff:CalculateDurationAndTickTime()
	-- call this after setup
	--TODO: maybe automate this somehow
	--TODO: only apply haste for certain buffs	
	--TODO: this probably crashes if we call it on a non-ticking buff	
	-- TODO: some system like we do for channel where we specify a number of ticks
	-- TODO: channel system is somewhat broken in that regard; no partial tick support there either
	
	if(self.duration~=nil)then
		local duration=self.duration
		if(self.durationaffectedbyhaste==true)then
			local hastemultiplier=self.combat:GetHasteCastReductionMultiplier()
			duration=duration*hastemultiplier
		end
		self.expirytime=self.starttime+duration
	end
	if(self.tickonbuffstart==true)then
		self:OnBuffTick(self.starttime)
		--self.totalticks=self.totalticks-1		
	end
	local tf=self:GetTickFrequency()
	if(tf~=nil)then
		self.nextticktime=self.starttime+tf
	end
end

function TRAININGGROUNDS_Buff:GetTickFrequency()
	if(self.tickfrequencyaffectedbyhaste==true)then
		local hastemultiplier=self.combat:GetHasteCastReductionMultiplier()
		return self.tickfrequency*hastemultiplier
	else
		return self.tickfrequency
	end
end

function TRAININGGROUNDS_Buff:step(elapsed,currenttime)
	-- override
	-- always call super when overriding, because we'll eventually start calling tick() from here
	
	
	--TODO: if we're not already calling step before expiry check, do so
	-- otherwise we miss the last tick
	if(self.nextticktime~=nil)then
	
		if(currenttime>=self.nextticktime)then
			
			self:OnBuffTick(self.nextticktime)
			
			self.nextticktime=self.nextticktime+self:GetTickFrequency()
		end
		
	end
end

function TRAININGGROUNDS_Buff:OnReapply(existingbuff,executiontime)
	if(self.duration~=nil)then
		existingbuff.expirytime=executiontime+self.duration
	end
	--TODO: cleanup new (unused) buff and any visuals it happens to be using
end

function TRAININGGROUNDS_Buff:SelfDispel()	
	--override and call super
	self.dying=true
	--print("TRAININGGROUNDS_Buff:SelfDispel()")
end


function TRAININGGROUNDS_Buff:draw(currenttime)
	--print("Buff:draw currenttime: "..currenttime)	
	-- TODO: WARNING: buffs are drawn from topright instead of topleft
		-- make this more consistent with everything else
	if(self.frame~=nil) then
		--self.frame:SetPoint("TOPRIGHT",self.combat.character.environment.frame,"TOPRIGHT",self.x,self.y)
		self.frame:SetPoint("TOPRIGHT",self.combat.character.environment.game.environment_ui.frame,"TOPRIGHT",self.x,self.y)		
		local durationtext
		if(self.expirytime~=nil)then
			durationtext=floor((self.expirytime-currenttime)/1000)
		else
			durationtext="" --TODO: hide fontstring
		end
		if(durationtext~="" and durationtext<3) then
			durationtext=_TRAININGGROUNDS:round((self.expirytime-currenttime)/1000,1)
		end
		self.frame.fontstring:SetText(durationtext)
		self.frame.fontstring:Show()
		----TODO: 
		--self.frame:SetFrameLevel(255)
		self.frame:Show()
	end
	
	if(self.stacksframe~=nil)then
		if(self.stacks~=nil and self.stacks>1)then
		--if(self.stacks~=nil and self.stacks>0)then  --!!!
		--if(true)then		--!!!			

			--self.stacksframe:ClearAllPoints()
			--self.stacksframe:SetAllPoints()
			--self.stacksframe:SetPoint("BOTTOMRIGHT",self.frame,"BOTTOMRIGHT",0,0)
			-- stacksframe is 32x32
			-- TODO: stacks fontstring text should be aligned right+bottom
			self.stacksframe:SetPoint("BOTTOMRIGHT",self.frame,"BOTTOMRIGHT",8,-8)
			self.stacksframe:Show()
			self.stacksframe.fontstring:Show()
			self.stacksframe.fontstring:SetText(self.stacks)
			-- self.frame.fontstring:Show()
			-- self.frame.fontstring:SetText(self.stacks)

			
		else
			self.stacksframe:Hide()
		end
	end
end

-- WARNING: OnBuffStart is called while iterating through the list of newbuffs
-- 				Don't create more buffs from this function!
function TRAININGGROUNDS_Buff:OnBuffStart()	
	--override
end


function TRAININGGROUNDS_Buff:OnBuffTick(ticktime)	
	--override
end

--TODO: differentiate between buff timing out vs early dispel
function TRAININGGROUNDS_Buff:OnBuffEnd(endtime)
	--override.  call super when overriding!
	--TODO: partial tick
	self:Cleanup()
end

--TODO: OnOwnerTakeDamage
-- TakeDamage is for unmitigated damage taken that we might want to mitigate
-- RecordDamage is the final damage taken recorded in the hypothetical combat log

function TRAININGGROUNDS_Buff:OnOwnerRecordDamage(damage,enemycharacter,spell,damagetime)
	--override
end

function TRAININGGROUNDS_Buff:OnOwnerDie(damagetime)
	--override
end

function TRAININGGROUNDS_Buff:Cleanup()
	if(self.frame~=nil) then
		--TRAININGGROUNDS_ReusableFrames.removeFrame(self.frame)
		self.frame:Cleanup()
	end
	if(self.stacksframe~=nil) then
		--TRAININGGROUNDS_ReusableFrames.removeFrame(self.stacksframe)
		self.stacksframe:Cleanup()
	end	
	self.spell=nil
end