
TRAININGGROUNDS_CombatModule = {}
TRAININGGROUNDS_CombatModule.__index = TRAININGGROUNDS_CombatModule

function TRAININGGROUNDS_CombatModule:new()
	local self={}
	setmetatable(self, TRAININGGROUNDS_CombatModule)	

	return self
end

TRAININGGROUNDS_CombatModule.lagtolerance=400
--TRAININGGROUNDS_CombatModule.lagtolerance=800


function TRAININGGROUNDS_CombatModule:setup(character)
		
	self.character=character
	self.target=nil
	-- targetchangecount is mostly for the benefit of the tutorial
	self.targetchangecount=0

	self.spellqueue={}
	self.castbarqueuedspell=nil
	self.castbarqueuedspelltick=0
	self.cooldownqueuedspell=nil
	self.cooldownqueuedspelltick=0
	self.movementqueuedspell=nil
	
	self.spellbook={}
	self.spellbuttons={}

	local spellbutton;

	self.resources={}

	self.hp=TRAININGGROUNDS_Resource.new()
	self.hp:setup(self,"hp",1,100,100,0)
	self.hp.visibleinnameplate=false	--!!!
	tinsert(self.resources,self.hp)
	
	-- set up resources before applying spells.
	-- spells will link back to these resources
	
	self.hastepercent=0
	self.masterypercent=0
	
	self:SetResources()
	self:SetSpells()
	
		
	self.newbuffs={}
	self.buffs={}
	
	self.channeledspell=nil
	self.nextchannelticktime=nil
	self.channelticksremaining=nil
	
	self.castedspell=nil
	-- Spell keeps track of this too, but we have to keep track of it here
		-- to know precisely when a castbarqueued spell is executed
		-- (by the time we need to know when, self.castedspell will be nil)
	self.castendtime=0
	
	
	self.previousspell=nil
	self.previousgcdspell=nil
	self.previous2gcdspell=nil	
	self.combocount=0
	self.idlecount=0
	
	-- cheap workaround until we can get Combo Strikes to snapshot
	self.lastattackwasokforcombostrikes=false
	
	
	--TODO: split Combo Strikes off to be WW only
	-- but have some sort of base WW class that the others can inherit
		-- (that base class might look remarkably like what CombatModule looks like now)
		
	-- Combo Strikes currently runs off the "previousgcdspell" system
	-- eventually we need to move it to its own system which checks whether AOE spells actually hit anything
	-- however, that requires each spellcast+buff to be given its own instance, and we don't have such a system yet
	

	
	self.lastexecutedspell=nil
	self.lastexecutedspelltime=0
	

	-- This is a list of delayed damage and special effects
	-- from various slow-acting spells and abilities.
	-- Format: impact.tick=number, impact.function=function
	-- QueueImpact(tick, function)
	self.impacteffects={}

	
	-- currently hiding HP bars until damage is taken
	self.takendamage=false
	
	self.ai=nil
	
	
	--for statistics:
	self.totalspellusecount=0	
	self.combatstarttime=nil
	self.debugspelltexts={}
	self.uptime=0
	
	self.pets={}
	
	self.spellerrortextoffcooldowntick=0
end

function TRAININGGROUNDS_CombatModule:SetResources()
	-- override
end
function TRAININGGROUNDS_CombatModule:SetSpells()
	-- override
end


-- you can call this function again at any point to redraw the nameplate.
--TODO: actually check that it works the way we say it does
	-- (in particular even if it successfully cleans up the old nameplate,
		-- we're not sure current resource levels will be drawn correctly)
function TRAININGGROUNDS_CombatModule:CreateNameplate()

	if(self.nameplate~=nil)then
		self.nameplate:Cleanup()
	end
	-- self.character.nameplate=TRAININGGROUNDS_Nameplate.new()
	-- self.character.nameplate:setup(self)--,20,self.character.modelinfo.visibleheight)

	self.nameplate=TRAININGGROUNDS_Nameplate.new()

	self.nameplate:setup(self)--,20,self.character.modelinfo.visibleheight)	
end


function TRAININGGROUNDS_CombatModule:CreateVisibleFrames()
	--TODO: some of this next block can be moved to resource:setup
	local r=0
	for i,v in ipairs(self.resources) do		
		v.frame:SetWidth(200);v.frame:SetHeight(25)
		v.frame.parent=_TRAININGGROUNDS.game.environment_ui.frame
		v.frame:SetPoint("TOPLEFT",v.frame.parent,"TOPLEFT",10,r*-25)
		--v.frame.framelevel=self.character.environment.frame.framelevel+1
		v.frame.framelevel=v.frame.parent.framelevel+1		
		v.frame:SetParent(v.frame.parent)
		v.frame:SetFrameLevel(v.frame.framelevel)
		if(_TRAININGGROUNDS.debug_resourcetextboxes==true)then
			v.frame:Show()
		end
		
		
		v.frame.fontstring:SetJustifyH("LEFT")
		v.frame.fontstring:SetText("Hello World "..v.name)
		v.frame.fontstring:Show()
		
		r=r+1
	end
	
	-- self.channelbar={}
	-- self.channelbar.frame=TRAININGGROUNDS_ReusableFrames.getFrame()
	-- self.channelbar.frame:SetWidth(200);self.channelbar.frame:SetHeight(25)
	-- self.channelbar.frame.parent=_TRAININGGROUNDS.game.environment_ui.frame
	-- self.channelbar.frame:SetPoint("TOPLEFT",self.channelbar.frame.parent,"TOPLEFT",10,r*-25)	
	-- self.channelbar.frame:SetParent(self.channelbar.frame.parent)
	-- self.channelbar.frame.framelevel=self.character.environment.frame.framelevel+1
	-- self.channelbar.frame:SetFrameLevel(self.channelbar.frame.framelevel)
	-- if(_TRAININGGROUNDS.debug_resourcetextboxes==true)then
		-- self.channelbar.frame:Show()			
	-- end		
	-- self.channelbar.frame.fontstring:SetJustifyH("LEFT")
	-- self.channelbar.frame.fontstring:SetText("")
	-- self.channelbar.frame.fontstring:Show()	
	

	
	
	
end




function TRAININGGROUNDS_CombatModule:CreateSpellButtons()
	-- override 
end

function TRAININGGROUNDS_CombatModule:CreateActionBar()
	-- private-ish function, called from CreateSpellButtons child functions
	-- call only once, unless you cleanup the action bar from somewhere else
	
	
	self.actionbar={}
	self.actionbar.frame=TRAININGGROUNDS_ReusableFrames.getFrame()
	self.actionbar.frame.parent=_TRAININGGROUNDS.game.environment_ui.frame
	self.actionbar.frame:SetWidth(640);self.actionbar.frame:SetHeight(96);
	self.actionbar.frame:SetPoint("BOTTOMLEFT",self.actionbar.frame.parent,"BOTTOMLEFT",0,16)
	--self.actionbar.frame.texture:Show()
	self.actionbar.frame:Show()
	--self.actionbar.frame.framelevel=self.character.environment.frame.framelevel+1
	--!!!
	--TODO:
	self.actionbar.frame.framelevel=_TRAININGGROUNDS.game.environment_ui.frame.framelevel+1
	--print(self.character.environment.frame,self.character.environment.frame.framelevel,self.actionbar.frame.framelevel)
	self.actionbar.frame:SetFrameLevel(self.actionbar.frame.framelevel)
	self.actionbar.frame:SetParent(self.actionbar.frame.parent)
end



function TRAININGGROUNDS_CombatModule:SetMaximumAndCurrentHP(hp)
	self.hp.max=hp
	self.hp.value=hp
end


function TRAININGGROUNDS_CombatModule:step(elapsed)



	local currenttime=self.character.environment.currenttime
	local i
	
	--TODO: call ChangeTarget so we update targetchangecount?
	if(self.target~=nil)then
		if(self.target.dead==true)then
			self.target=nil
		end
	end
	
	-- check if we've been hit by any delayed effects
	for i=#self.impacteffects,1,-1 do
		local impact=self.impacteffects[i]
		if(impact.tick<=currenttime)then
			impact.fn(impact.tick)
			tremove(self.impacteffects,i)
		end
	end
	

	-- buffs
	-- add newbuffs to buffs list
	for i=1,#self.newbuffs,1 do
		tinsert(self.buffs,self.newbuffs[i])
		self.newbuffs[i]:OnBuffStart(self.newbuffs[i].starttime)
	end
	self.newbuffs={}
	
	for i=1,#self.buffs,1 do
		self.buffs[i]:step(elapsed,currenttime)
	end
	
	for i=#self.buffs,1,-1 do
		if((self.buffs[i].expirytime~=nil and currenttime>self.buffs[i].expirytime) or self.buffs[i].dying==true)then
			self.buffs[i].dead=true
			self.buffs[i]:OnBuffEnd(self.buffs[i].expirytime)
			tremove(self.buffs,i)
		end
	end
	
	-- resource step (regen, etc) ---------------------------------
	for i=1,#self.resources,1 do
		self.resources[i]:step(elapsed)
	end
	
	-- spell recharges --------------------------------------------
	for i=1,#self.spellbook,1 do
		local spell=self.spellbook[i]
		local oldrechargetick=spell.nextrechargetick
		if(spell.charges<spell.maxcharges)then
			if(currenttime>=spell.nextrechargetick)then
				spell.charges=spell.charges+1
				if(spell.charges>=spell.maxcharges)then
					spell.charges=spell.maxcharges		
				else
					spell.nextrechargetick=spell.nextrechargetick+spell:GetRechargeForUse()
					self:DrawOneCooldown(spell,oldrechargetick)
				end
			end
		end
	end
	
	-- spell cooldown-into-recharge animations --------------------
	for i=1,#self.spellbook,1 do
		local spell=self.spellbook[i]
		--TODO: avoid doing this every frame
		if(currenttime>=spell.offcooldowntick)then
			if(spell.charges<spell.maxcharges)then
				self:DrawOneCooldown(spell,spell.offcooldowntick)
			end
		end
	end
	
	
	
	-- -- uptime -----------------------------------------------------
	
	-- --TODO: this doesn't work if the channel is interrupted
	-- if(self.channeledspell~=nil)then
		-- if(self.channeledspell.channelendtime>currenttime)then
			-- self.uptime=self.uptime+elapsed
		-- else
			-- if(self.channeledspell.channelendtime>currenttime-elapsed)then
				-- self.uptime=self.uptime+(currenttime-self.channeledspell.channelendtime)
			-- end
		-- end
	-- end
	
	-- ---------------------------------------------------------------
	
	-- castbars ------------------------------------------------------
	
	if(self.castedspell~=nil)then
		if(currenttime>=self.castedspell.castendtime)then
			--print("CombatModule:step -- finished casting; do something")
			--TODO: must check spell requirements again here
			
			-- remember castendtime in case there's a castbarqueued spell
			self.castendtime=self.castedspell.castendtime			
			self.castedspell:executeSuccessfullyCastEnd(self.castedspell.castendtime)
		end		
	end

	if(self.castedspell~=nil)then
		--TODO: better "movement" check,
		--and some buffs allow casting while moving
		if(not
		(self.castedspell.castablewhilemoving==true or 
		(self.character.xspeed==0 and self.character.yspeed==0 and self.character.zspeed==0 and self.character.z<=0))
		)then
			self.castendtime=currenttime
			self.castedspell:StopCastingOrChannelingOldSpell(self.castendtime)
		end
	end
	
	---------------------------------------------------------------
	
	
	-- channels ---------------------------------------------------
	if(self.channeledspell~=nil)then
		--self.channelbar.frame.fontstring:SetText(self.channeledspell.spellname..": ".._TRAININGGROUNDS:round((self.channeledspell.channelendtime-currenttime)/1000,1))
		--print("CombatModule.step channels")
		
		--TODO: this can potentially miss a tick if framerate dies at channelendtime
		if(currenttime>=self.nextchannelticktime)then			
			self.channeledspell:OnChannelTick(self.nextchannelticktime)
			self.nextchannelticktime=self.nextchannelticktime+self.channeltickfrequency
			self.channelticksremaining=self.channelticksremaining-1	
		end
		
		if(currenttime>=self.channeledspell.channelendtime)then
			self:StopChanneling(currenttime)
		end
	end
	---------------------------------------------------------------
	
	
	
	
	-- spellcasts ----------------------------------------------------------------------------
	
	-- first deal with anything in the movement queue
	-- Automove is intended to work only when the spell in question is otherwise 100% available to cast if the player were in range.
	-- It won't (shouldn't) work if the player has insufficient resources or if cooldown (GCD or otherwise) is active.
	
	--TODO: this check should be redundant, but probably isn't redundant as currently implemented
	if(self.movementqueuedspell~=nil)then
		if(self.movementqueuedtarget==nil)then
			self.movementqueuedspell=nil
		end
		if(self.target~=nil)then
			if(self.target~=self.movementqueuedtarget)then
				self.movementqueuedspell=nil
			end
			if(self.target.combat~=nil)then
				if(self.target.combat.hp.value<=0)then
					self.movementqueuedspell=nil
				end
			end
		else	
			self.movementqueuedspell=nil
		end
	end
	--TODO: we hastily fixed the movementqueue-can't-cast-while-moving problem, might have broken something else in the process
	if(self.movementqueuedspell~=nil)then
		--TODO: single function for "is can't-cast-while-moving-spell castable" check
		if(self.movementqueuedspell.castablewhilemoving==true
		or (self.character.xspeed==0 and self.character.yspeed==0 and self.character.zspeed==0 and self.character.z<=0)
		)then
			-- clear the queue first
			-- if player is still out of range, spell will be placed back in queue
			local spell=self.movementqueuedspell
			self.movementqueuedspell=nil
			self:tryExecuteSpell(spell,currenttime)		
		end
	end
	
	-- then deal with anything in the cooldown queue
	if(self.cooldownqueuedspell~=nil)then
										--3rd arg: true: cooldownqueued.
										-- execute spell time will be whenever spell last came off cooldown.
		self:tryExecuteSpell(self.cooldownqueuedspell,currenttime,true)	
	end	
	-- spells can only stay in the queue for (lagtolerance) ms
	-- (must wait to perform this check until AFTER trying to execute, or spell will expire first)
	if(self.cooldownqueuedspell~=nil)then
		if(currenttime>self.cooldownqueuedspelltick+TRAININGGROUNDS_CombatModule.lagtolerance)then			
			self.cooldownqueuedspell=nil	
		end
	end

	-- then check the castbar queue
	if(self.castbarqueuedspell~=nil)then
		-- TODO: there is a slight possibility that if we got the logic wrong at any point,
			-- self.castendtime is still zero and weird things will happen to the spell's cooldown
										-- 4th arg: true: castbarqueued.
										-- execute spell time will be whenever previous spell finished casting.
		--print("about to tryexec ",currenttime)
		self:tryExecuteSpell(self.castbarqueuedspell,currenttime,false,true)
	end	
	-- spells can only stay in the queue for (lagtolerance) ms
	-- (must wait to perform this check until AFTER trying to execute, or spell will expire first)
	if(self.castbarqueuedspell~=nil)then
		if(currenttime>self.castbarqueuedspelltick+TRAININGGROUNDS_CombatModule.lagtolerance)then			
			self.castbarqueuedspell=nil			
		end
	end
	
	
	
	-- next, attempt to execute all spells in the normal queue.
	-- if any of them are on cooldown, try to move them to the cooldown queue.
	-- limit 1 spell in the cooldown queue; newer spells overwrite older ones.
	for i,v in ipairs(self.spellqueue) do
		--print("CombatModule",self,"spell in queue")
		if(self.castedspell~=nil)then
			--TODO: cooldownqueue is currently doubling as castbar queue -- might want to make a separate queue for castbar		
			self.castbarqueuedspell=v
			self.castbarqueuedspelltick=currenttime
		else
			if(currenttime>=v.offcooldowntick)then
				--print("CombatModule",self,"cooldown OK")
				self:tryExecuteSpell(v,currenttime,false)
			else
				--TODO: the following is a cheap awkward workaround for FSK slam attack -- do something better
				--don't use this for anything other than FSK
					-- the game will still try to queue the spell after calling this function
				-- ... we'll probably use a "tornado kick" style spellbutton swap
				
				local executeddespitecooldown=v:executeDespiteCooldown(currenttime)
				--print("EDD: ",executeddespitecooldown)

				
				-- In order to place a spell in the cooldown queue, we must already have the required resources at the time the spell is queued.
				--		(It is still required to check if those resources still exist when the spell comes off cooldown.)
				local haveresources=v:haveResources()
				if(haveresources.success==true)then
					-- However, it is still possible to queue a spell still on cooldown as long as it will come off cooldown in (lagtolerance) ms.			
					if(currenttime>v.offcooldowntick-TRAININGGROUNDS_CombatModule.lagtolerance)then
						--print("CombatModule.lua cooldownqueuing "..v.spellname.." at "..currenttime)
						if(v.queueable==true)then
							--print("Spell to cooldownqueue: ",v)
							self.cooldownqueuedspell=v
							self.cooldownqueuedspelltick=currenttime
						end
					else
						--!!!
						if(executeddespitecooldown~=true)then
							self:CreateSpellErrorText("Spell is not ready yet.")
						end
					end
				else					
					--else don't have resources; display spell error message
					self:CreateSpellErrorText(haveresources.message)
				end
			end
		end
	end
	self.spellqueue={}
	
	-- end spellcasts block -------------------------------------------------------------------
	
	
	-- approximate uptime check
	if(self.combatstarttime~=nil)then
		if(self.gcdtracker~=nil)then
			if(self.gcdtracker.offcooldowntick>currenttime
			or self.channeledspell~=nil)then			
				self.uptime=self.uptime+elapsed*1000
			end
		end
	end
	

	-- downtime check		
	if(self.comboresettime~=nil) then
		if(self.channeledspell==nil)then
			if(currenttime>self.comboresettime)then
				self.comboresettime=self.comboresettime+1000
				self.combocount=0
				self.idlecount=self.idlecount+1
				if(_TRAININGGROUNDS.debug_showdowntimeboxes==true)then
					local idletext="|cffff0000nothing"
					if(self.idlecount>1) then
						idletext=idletext.." x"..self.idlecount
					end
					idletext=idletext.."..."
					self.character.environment.game:createSpellTextbox(idletext)
				end
			end
		end
	end
	
	
	if(self.ai~=nil)then
		self.ai:step(elapsed,currenttime)
	end
	if(self.nameplate~=nil)then
		self.nameplate:step(elapsed)
	end
	
	for i=1,#self.spellbook,1 do
		local spell=self.spellbook[i]
		spell:UpdateButtonChargeDisplay(currenttime)
	end
	
	if(true) then
		--TODO: debug to turn flash on/off				
		if(self.lastexecutedspell~=nil)then
			if(currenttime>self.lastexecutedspelltime+0)then
				if(self.lastexecutedspell.button~=nil)then
					ActionButton_HideOverlayGlow(self.lastexecutedspell.button.frame)
				end
				self.lastexecutedspell=nil			
			end
		end
	end
	
	if(self.totaltimetextbox~=nil)then 		
		if(self.combatstarttime~=nil)then								
			local totaltime=(currenttime-self.combatstarttime)
			local totaltimeseconds=totaltime/1000.0
			--SOURCE: stackoverflow.com/questions/17474312/using-lua-to-format-0-seconds-as-000000
			local totaltimestring=string.format("%.2d:%.2d:%.2d", totaltimeseconds/60%60, totaltimeseconds%60, (totaltime/10)%100)
			self.totaltimetextbox.frame.fontstring:SetText("Total time: "..totaltimestring)			
		end
	end
	if(self.apmtextbox~=nil)then
		if(self.combatstarttime~=nil and self.combatstarttime>=0)then
			local totaltime=(currenttime-self.combatstarttime)
			local minutes=(totaltime/60000)
				--TODO: rename to self.totalactions?
			local apm=(self.totalspellusecount/minutes)
			if(apm>999)then apm=999 end
			local apmstr=round(apm,1)
			self.apmtextbox.frame.fontstring:SetText("Actions per minute: "..apmstr)
		end
	end
	if(self.uptimetextbox~=nil)then
		if(self.combatstarttime~=nil)then								
			local totaltime=(currenttime-self.combatstarttime)
			--print(self.uptime,totaltime)
			local uptimepct=self.uptime/totaltime*100
			local uptimestr=round(uptimepct,1)
			if(uptimepct<10)then
				uptimestr="0"..uptimestr
			end			
			if(uptimepct<=0)then
				uptimestr="00.0"
			end			
			self.uptimetextbox.frame.fontstring:SetText("Uptime: "..uptimestr.."%")	
		end		
	end
	
	for i=#self.pets,1,-1 do
		if(self.pets[i].dead==true)then
			tremove(self.pets,i)
		end
	end	
	
end



function TRAININGGROUNDS_CombatModule:draw(currenttime)
	-- draw resources
	for i=1,#self.resources,1 do	
		self.resources[i]:draw(currenttime)
	end

	-- draw buffs
	local r=1
	for i=1,#self.buffs,1 do	
		self.buffs[i].x=-r*40
		self.buffs[i].y=-8
		--print("CombatModule:step drawbuffs "..self.buffs[i].x..","..self.buffs[i].y)
		self.buffs[i]:draw(currenttime)
		if(self.buffs[i].visibletoplayer) then
			r=r+1
		end
	end
	
	-- draw spellbuttons based on available resources
	for i=1,#self.spellbuttons,1 do
		local button=self.spellbuttons[i]
		local spell=self.spellbuttons[i].spell
		if(spell~=nil) then
			if(spell:haveResources().success==true)then
				button.frame:SetAlpha(1.0)
			else				
				button.frame:SetAlpha(0.5)
			end
		end
	end
end



function TRAININGGROUNDS_CombatModule:tryQueueSpell(spell)
	--local currenttime=GetTime()*1000
	--print("CombatModule:tryQueueSpell "..spell.spellname)
	if(spell==self.blackoutkick and spell.istornado==true) then
		--print("CombatModule:tryQueueSpell switch to tornadokick")
		tinsert(self.spellqueue,self.tornadokick)
	else
		tinsert(self.spellqueue,spell)			
	end
	

end

function TRAININGGROUNDS_CombatModule:TryMimicSpell(playerspell,currenttime)
	--override
end

-- if this function is called with cooldownqueued==true,
-- the execute spell tick will be whenever that spell came off cooldown...if that's in the distant past, it could break things.
function TRAININGGROUNDS_CombatModule:tryExecuteSpell(spell,currenttime,cooldownqueued,castbarqueued)
	local result=false
	
	if(self.character.deadingame==true)then
		return result
	end

	local r=spell:haveResources()
--print("CombatModule ",self,"!!!")
	
	--TODO: reorder checks to match WoW's error message priority
	
	--TODO: buff to allow cast while moving
	--TODO: better check for "character is moving"
	if(spell.castablewhilemoving==true or 
	(self.character.xspeed==0 and self.character.yspeed==0 and self.character.zspeed==0 and self.character.z<=0))then
		--TODO: some spells can be used while casting other spells
		if(self.castedspell==nil)then
			local haveresources=spell:haveResources()
			if(haveresources.success==true)then
		--print("CombatModule ",self,"???")	
				--TODO: move to Spell:tryExecute
				if(currenttime>=spell.offcooldowntick)then		
					--print("tryExec ",currenttime,spell.offcooldowntick)
					--TODO: split target and range checks into separate functions
					local tryagain=true
					local loops=1
		--print("CombatModule ",self,"===")
					-- Roll/FSK/SCK pacify check
					-- Check if buffs exist
					
					local roll=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_Roll)
					local fsk=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_FlyingSerpentKick)
					local sck=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_SpinningCraneKick)
					
					
					-- If roll exists, and time remaining is >gracetime, attempt fails
							-- (gracetime could be 0 ms)
					-- If FSK exists, and spell isn't FSK, attempt fails
					
					if(roll==nil and (fsk==nil or spell.name=="Flying Serpent Kick") and sck==nil)then
					
						while(tryagain==true and loops<=2)do
							tryagain=false
										
							if(spell:TargetIsOkay(self.character,self.target).success==true)then
								if(spell:RangeIsOkay(self.character,self.target).success==true)then
									--TODO: check if there a channeled spell preventing any other spell from being cast
									--	(SCK blocks other spells during channel, FoF doesn't)
									--TODO: stop any existing channel if spellcast isn't prevented
									
									local executespelltick=currenttime
									if(cooldownqueued==true and spell.offcooldowntick<currenttime)then
										executespelltick=spell.offcooldowntick
									end
									if(castbarqueued==true and self.castendtime<currenttime)then
										executespelltick=self.castendtime
									end
									if(_TRAININGGROUNDS.debug_printabilityexecutetime==true)then
										print("CombatModule.lua casting "..spell.spellname.." at "..executespelltick)
									end
									
									-- success!
									result=true
									spell:executeSuccessfullyCastStart(executespelltick)					
									self.movementqueuedspell=nil
									self.cooldownqueuedspell=nil
									if(true)then
										--TODO: debug to turn flash on/off				
										if(self.lastexecutedspell~=nil)then									
											if(self.lastexecutedspell.button~=nil) then
												ActionButton_HideOverlayGlow(self.lastexecutedspell.button.frame)
											end
										end					
										self.lastexecutedspell=spell
										self.lastexecutedspelltime=currenttime
										if(self.ai~=nil)then
											self.ai.lastexecutedspelltime=currenttime
										end
										if(spell.button~=nil)then
											ActionButton_ShowOverlayGlow(spell.button.frame)									
										end
									end
									--!!!
									--self.character.combatstanceendtick=currenttime+3000 	-- use this line for what we think of as standard combat
									self.character.combatstanceendtick=currenttime+5000		-- use this line for the W7 monk test
								else
									--TODO: else out of range; display spell error message	
									
									-- if this was a range issue, and we were otherwise OK to cast, place spell in movement queue and start automove
									--TODO: separate range from targetOK check so we don't try to automove into a dead target
									if(spell:RangeIsOkay(self.character,self.target).success==false)then
										self.movementqueuedspell=spell
										self.movementqueuedtarget=self.target
									end
								end
							else
								--TODO: else invalid target; display spell error message
								
								-- attempt to autotarget closest enemy
								self:Autotarget()
								if(self.target~=nil)then
									tryagain=true
									loops=loops+1
								end
								
							end
						end
					else
						-- else pacified from roll or FSK
					end
				else
					--TODO: else on cooldown; display spell error message
					--TODO: it is entirely possible that we'll never see this error message
					-- because the spell expires in queue before this function is ever called
					--self:CreateSpellErrorText("Spell is not ready yet")					
					-- ...current situation:
					-- we're calling this from the queue check
					-- and also checking for executeDespiteCooldown
					-- so we don't get "SEF: Focus" overlapping with the error message
				end
			else	
				--TODO: else don't have resources; display spell error message
				self:CreateSpellErrorText(haveresources.message)
			end
		else
			--else can't use ability while casting a spell
			--TODO: should be able to queue new spell
		end
	else
		--else can't cast while moving
	end
	return result
end


function TRAININGGROUNDS_CombatModule:CreateSpellErrorText(errortext)	
	-- TODO: override instead of isa
	if(self.character:isa(TRAININGGROUNDS_PlayerCharacter))then
		if(self.gcdtracker.offcooldowntick<=self.character.environment.currenttime)then
			if(self.spellerrortextoffcooldowntick<=self.character.environment.currenttime)then
				self.spellerrortextoffcooldowntick=self.character.environment.currenttime+1000
				errortext="|cFFFF0000("..errortext..")"
				_TRAININGGROUNDS.game:createSpellTextbox(errortext)
			end
		end
	end
end


function TRAININGGROUNDS_CombatModule:Autotarget()
	--TODO: separate tabtarget function that cycles through different enemies
	--	(...or not, given how often tabtarget doesn't cooperate in WoW)
	
	local MAX_DISTANCE=300
	local environment=self.character.environment
	local bestdistancesqr=nil
	local target=nil
	for i=1,#environment.gameobjects,1 do
		local obj=environment.gameobjects[i]
		--if(obj~=self.character)then
		--!!!
		if(obj:isa(TRAININGGROUNDS_Enemy))then
			if(obj.combat~=nil)then
				if(obj.deadingame~=true)then
					--TODO: check if obj is hostile and targetable -- probably call TargetIsOkay
											-- might require passing current spell to autotarget
					local distancesqr=(self.character.x-obj.x)*(self.character.x-obj.x)+(self.character.y-obj.y)*(self.character.y-obj.y)
					if(distancesqr<=MAX_DISTANCE*MAX_DISTANCE)then
						if(bestdistancesqr==nil or bestdistancesqr>distancesqr)then
							target=obj
							bestdistancesqr=distancesqr
						end
					end
				end
			end
		end
	end
	if(target~=nil)then
		self:ChangeTarget(target)
	end
	
end


-- This will not find buffs in the newbuff list.
function TRAININGGROUNDS_CombatModule:GetBuffIfExists(buffclass)
	for i=1,#self.buffs,1 do
		--print("TRAININGGROUNDS_CombatModule checking if "..tostring(self.buffs[i]).." isa "..tostring(buffclass))
		if(self.buffs[i]:isa(buffclass))then
			if(self.buffs[i].dying~=true and self.buffs[i].dead~=true)then
				return self.buffs[i]			
			end
		end
	end
	return nil
end

-- This will ONLY find buffs in the newbuff list.
function TRAININGGROUNDS_CombatModule:GetNewBuffIfExists(buffclass)
	for i=1,#self.newbuffs,1 do
		--print("TRAININGGROUNDS_CombatModule checking if "..tostring(self.buffs[i]).." isa "..tostring(buffclass))
		if(self.newbuffs[i]:isa(buffclass))then
			if(self.newbuffs[i].dying~=true and self.newbuffs[i].dead~=true)then
				return self.newbuffs[i]			
			end
		end
	end
	return nil
end


-- do not call the OnApplyCooldown event from here
-- ...that event is specifically for non-GCD cooldowns
function TRAININGGROUNDS_CombatModule:setGCD(executetime,duration)
	--print("setGCD",executetime,duration)
	for i=1,#self.spellbook,1 do
		local spell=self.spellbook[i]
		if(spell.usesgcd)then
			if(spell.offcooldowntick<executetime+duration)then
				spell.offcooldowntick=executetime+duration
				spell.oncooldowntick=executetime
			end
		end
	end
	self.comboresettime=executetime+1000
end

function TRAININGGROUNDS_CombatModule:SetComboCount(executetime,spell)
	--print("CombatModule:SetComboCount "..tostring(self.previousgcdspell)..","..tostring(spell))
	if(self.previousgcdspell~=spell) then
		self.combocount=self.combocount+1		
	else
		self.combocount=0
	end
	--if(self.combocount>2) then
	if(self.combocount>0) then
		if(_TRAININGGROUNDS.debug_showcomboboxes==true)then
			self.character.environment.game:CreateComboTextbox(self.combocount)
		end
	end
	self.idlecount=0
end

function TRAININGGROUNDS_CombatModule:DrawAllCooldowns(currenttime)	
	-- TODO: game wants oncooldowntick to be WoW time, not local time
	for i=1,#self.spellbuttons,1 do		
		local button=self.spellbuttons[i]
		local spell=self.spellbuttons[i].spell
		self:DrawOneCooldown(spell,currenttime)
	end
end

function TRAININGGROUNDS_CombatModule:DrawOneCooldown(spell,currenttime)
	-- TODO: game wants oncooldowntick to be WoW time, not local time
	local button=spell.button
	if(button~=nil)then
		--TODO: it looks like spells with charges have SetDrawEdge(true) even if they have 0 charges left and are on cooldown
		--TODO: spell.onlyshowrechargenevercooldown logic is wrongish
			-- -- breaks for spells with 1 maxcharge and no cooldown
		if(spell.offcooldowntick>currenttime and (not spell.onlyshowrechargenevercooldown))then
			local onCDtick=self.character.environment.game:ConvertLocalTimeToRealTime(spell.oncooldowntick,self.character.environment)
			local offCDtick=self.character.environment.game:ConvertLocalTimeToRealTime(spell.offcooldowntick,self.character.environment)
			
			--print("TRAININGGROUNDS_CombatModule:DrawAllCooldowns() "..tostring(spell).." "..spell.oncooldowntick.." "..spell.offcooldowntick)
			--if(spell.maxcharges==1)then
			button.frame.cooldown:SetDrawEdge(false)
			--end
			button.frame.cooldown:SetCooldown(onCDtick/1000.0,(offCDtick-onCDtick)/1000.0)
			--button.frame.cooldown:SetCooldown(GetTime(),10)
			
			--ActionButton_ShowOverlayGlow(button.frame)
			--ActionButton_HideOverlayGlow(button.frame)
			
			--ActionButton_ShowOverlayGlow(PlayerFrame)

		else	-- else drawing recharge, not cooldown
			--if(spell==self.tigereyebrew)then	print(spell.charges,spell.maxcharges) end
			if(spell.charges<spell.maxcharges)then
				
																							--TODO: call function instead of defaultrecharge
																							-- and the logic here is probably wrong
				local onCDtick=self.character.environment.game:ConvertLocalTimeToRealTime(spell.nextrechargetick-spell.defaultrecharge,self.character.environment)
				local offCDtick=self.character.environment.game:ConvertLocalTimeToRealTime(spell.nextrechargetick,self.character.environment)
				--print(onCDtick,offCDtick)
				button.frame.cooldown:SetDrawEdge(true)
				
				-- if onCDtick==offCDtick (e.g. if cast time is greater than cooldown),
				-- the cooldown swirl malfunctions
				-- ...to fix it, we just set onCDtick to before offCDtick
				if(offCDtick<=onCDtick)then
					onCDtick=offCDtick-1
				end
				button.frame.cooldown:SetCooldown(onCDtick/1000.0,(offCDtick-onCDtick)/1000.0)
			end
			
		end
	end
end

function TRAININGGROUNDS_CombatModule:StopCasting(currenttime)
	if(self.castedspell~=nil)then
		self.castedspell:OnCastStop(currenttime)
		self.castedspell.caststarttime=nil
		self.castedspell.castendtime=nil
		self.castedspell=nil
	end
end

function TRAININGGROUNDS_CombatModule:StopChanneling(currenttime)
	if(self.channeledspell~=nil)then
		self.channeledspell:OnChannelStop(currenttime)
		self.channeledspell.channelstarttime=nil
		self.channeledspell.channelendtime=nil
		self.channeledspell=nil
		--self.channelbar.frame.fontstring:SetText("")
	end
end

function TRAININGGROUNDS_CombatModule:GetHasteCastReductionMultiplier()
	-- TODO: NYI
	local result=1/(1+self:GetTotalHastePercent())
	--print("CombatModule:GetHasteCastReductionMultiplier "..result)
	return result
	
	-- 50% haste: 2/3 multiplier
	
	-- need a multiplier below 0.8333... to fit two FoFs into a Serenity
	-- 1/(1+haste)=0.8333
	-- 1 = 0.8333+0.8333*haste
	-- 0.1666... = 0.8333*haste
	--  = haste = 0.2
end

function TRAININGGROUNDS_CombatModule:GetTotalHastePercent()
	-- TODO: NYI.  Does not affect energy regen (yet), but does affect channel time.
	--return self.hasterating/9000.0
	return self.hastepercent
	--return 0.11
	--return 0.0
end




function TRAININGGROUNDS_CombatModule:ChangeTarget(newtarget)
	-- changing targets always cancels automove
	self.movementqueuedspell=nil
	
	if(self.target~=newtarget)then 
		self.targetchangecount=self.targetchangecount+1 
	end
	local oldtarget=self.target
	self.target=newtarget

	if(self.showtargetcircles==true)then	
		--TODO: only change shadows/targetcircles if this is the player's combat module
		if(oldtarget~=nil) then
			oldtarget.shadowframe:SetFrameLevel(self.character.environment.shadowframelevel)
			oldtarget.shadowframe.texture:SetTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\shadow.tga")
			oldtarget.shadowframe.texture:SetGradientAlpha("HORIZONTAL",1,1,1,1.0,1,1,1,1.0);		
		end		
		if(self.target~=nil)then
			self.target.shadowframe:SetFrameLevel(self.character.environment.targetcircleframelevel)
			self.target.shadowframe.texture:SetTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\targetingcircle.tga")
			self.target:SetTargetingCircleColor(self.character)		
			--self.target.shadowframe.texture:SetGradientAlpha("HORIZONTAL",1,0,0,1.0,1,0,0,1.0);
		end
	end
	
	for i=1,#self.pets,1 do
		if(self.pets[i].copyownerstarget==true)then
			self.pets[i].combat:ChangeTarget(newtarget)
		end
	end	
end

--takes place after target change.
--TODO: call ChangeTarget from inside OnClickTarget instead of from Environment_World
function TRAININGGROUNDS_CombatModule:OnClickTarget(newtarget)
	--not currently used
	-- pet target changes are being handled directly by ChangeTarget
	-- (because we can change targets without clicking on them)
end


function TRAININGGROUNDS_CombatModule:QueueImpact(tick,fn)
	local impact={}
	impact.tick=tick
	impact.fn=fn
	tinsert(self.impacteffects,impact)
end

function TRAININGGROUNDS_CombatModule:ApplyDamageToEnemy(damage,enemy,spell)
	--TODO: show debug warning if spell==nil
	
	--TODO: don't hardcode exceptions
	--if(spell~=nil and spell==self.touchofdeath)then
	--		--don't apply buffs for ToD (currently includes Gale Burst)
	--else
		-- local teb=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_TigereyeBrew)
		-- if(teb~=nil)then
			-- damage=damage*1.30
		-- end
		local serenity=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_Serenity)
		if(serenity~=nil)then
			damage=damage*1.40
		end		
		if(spell~=nil and spell.procscombostrikes==true and self.lastattackwasokforcombostrikes==true)then			
			damage=damage*(1+self.masterypercent)	
		end
		local hc=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_HitCombo)
		if(hc~=nil)then			
			damage=damage*1+0.01*(hc.stacks)
			--print("Hit Combo stacks:",hc.stacks,"multiplier:",1+0.01*(hc.stacks))
		end		
	--end
	
	--TODO: SEF checks owner's buffs instead of their own
	--TODO: SEF 45% check
	damage=damage*self:GetSEFMultiplier()
	
	
	
	if(enemy.combat:RollToBlock(self.character)==true)then
		--TODO: multiply by block amount
		damage=damage*0
	end
	
	--TODO: buffs, debuffs, crit damage, etc
	self:ApplyExactDamageToEnemy(damage,enemy,spell)
	
end

--TODO: don't hardcode SEF
function TRAININGGROUNDS_CombatModule:GetSEFMultiplier()
	local sef=self:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_StormEarthAndFire)
	if(sef~=nil)then
		return TRAININGGROUNDS_Spell_W7_WindwalkerMonk_TigereyeBrew.SEFMULTIPLIER
	else
		return 1.0
	end
end

--TODO: this is considered directionless damage
-- don't necessarily need to rewrite, but maybe rename
function TRAININGGROUNDS_CombatModule:ApplyExactDamageToEnemy(damage,enemy,spell)
	enemy.combat:TakeDamageFromEnemy(damage,self,spell)
end

function TRAININGGROUNDS_CombatModule:RollToBlock(enemy)
	--override
	
	return false
end


function TRAININGGROUNDS_CombatModule:TakeDamageFromEnemy(damage,enemycombat,spell)
	--print("Damage taken: ",damage)
	if(damage>0)then
		self:RevealNameplateHP()
	end
	
	
	local oldhp=self.hp.value	
	--TODO: buffs, debuffs, etc
	--TODO: armor physical damage reduction, etc
	self.hp:AddToValue(-damage)
	local newhp=self.hp.value
	if(newhp<oldhp)then
		local enemy=enemycombat.character
		local hurtbox=self.nameplate:CreateDamageParticle(oldhp,newhp)
		--TODO: adjust behavior of particle based on attack
		local angle=math.atan2(self.character.y-enemy.y,self.character.x-enemy.x)
		local spread=.5
		angle=angle+math.random()*spread-spread*0.5
		hurtbox.xspeed=math.cos(angle)*200
		hurtbox.yspeed=math.sin(angle)*200
		--hurtbox.zspeed=200
		hurtbox.zspeed=300
		--  hurtbox expirytime was already set to duration 1000... (TODO:)
		hurtbox.expirytime=hurtbox.expirytime+500
		hurtbox.gravity=-300*4
	end
	
	local i
	for i=1,#self.buffs,1 do
		--TODO: spell, 3rd argument
		--TODO: damagetime, 4th argument
		self.buffs[i]:OnOwnerRecordDamage(damage,enemycombat.character,spell,self.character.environment.currenttime)
	end
end

function TRAININGGROUNDS_CombatModule:RevealNameplateHP()
	if(self.takendamage~=true)then
		self.hp.visibleinnameplate=true
		self.takendamage=true
		self:CreateNameplate()
	end	
end

function TRAININGGROUNDS_CombatModule:RevealNameplateCastbar()
	if(self.castbarrevealed~=true)then
		self.castbarrevealed=true		
		self:CreateNameplate()
	end
	
end


function TRAININGGROUNDS_CombatModule:OnDie(deathtime)
	--TODO: anything else that needs doing
	
	self:StopCasting()
	self:StopChanneling()
	
	for i=1,#self.buffs,1 do
		self.buffs[i]:OnOwnerDie(deathtime)
	end
	-- 
	for i=#self.buffs,1,-1 do
		self.buffs[i].dead=true
		self.buffs[i]:OnBuffEnd(deathtime)
		tremove(self.buffs,i)
	end		
	
	
end

function TRAININGGROUNDS_CombatModule:ClearAllSpells()
	for i=1,#self.spellbook,1 do
		local spell=self.spellbook[i]
		spell:Cleanup()
	end
	self.spellbook={}
end

function TRAININGGROUNDS_CombatModule:ClearAllSpellButtons()
	for i=1,#self.spellbuttons,1 do
		local spellbutton=self.spellbuttons[i]
		spellbutton:Cleanup()
		spellbutton.spell.button=nil
	end
	self.spellbuttons={}
	
	if(self.actionbar~=nil)then
		self.actionbar.frame:Cleanup()
		self.actionbar=nil
	end
end

function TRAININGGROUNDS_CombatModule:Cleanup()
	local i
	--print("TRAININGGROUNDS_CombatModule:Cleanup()")
	if(self.nameplate~=nil)then
		self.nameplate:Cleanup()
	end
	
	self:ClearAllSpells()
	self:ClearAllSpellButtons()
	
	--TODO: does this make a difference?
	self.impacteffects={}

	-- TODO: there are more used frames we haven't cleaned up yet
	-- TODO: call cleanup on child objects like resources, buffs, etc

	-- need to clean up buffs or some particle effects associated with them won't disappear properly
	for i=#self.buffs,1,-1 do
		self.buffs[i].dead=true
		self.buffs[i]:OnBuffEnd(self.buffs[i].expirytime)
		tremove(self.buffs,i)
	end	
	
	self.pets={}	
	
	-- TODO: call cleanup on channelbar (do we already do that in nameplate:cleanup?)

--!!!
-- these are gameobjects so they should be getting cleaned up automatically

	-- for i=1,#self.debugspelltexts,1 do
		-- self.debugspelltexts[i]:Cleanup()		
	-- end
	-- if(_TRAININGGROUNDS.debug_debugwindow==true)then
		-- self.totaltimetextbox:Cleanup()
		-- self.apmtextbox:Cleanup()
		-- self.uptimetextbox:Cleanup()	--TODO: uptime box still isn't getting cleaned up properly
	-- end
end