TRAININGGROUNDS_Nameplate = {}
TRAININGGROUNDS_Nameplate.__index = TRAININGGROUNDS_Nameplate

TRAININGGROUNDS_Nameplate.NAMEPLATE_HEIGHT=8

-- TODO: nameplates not part of clickarea yet

function TRAININGGROUNDS_Nameplate.new()
	local self=setmetatable({}, TRAININGGROUNDS_Nameplate)	
	
	return self
end

--yoffset is in screen coordinates
--zoffset is in camera coordinates (includes zstretch and zoom)
-- (use zoffset roughly equal to the mob's height, and a small yoffset on top)
function TRAININGGROUNDS_Nameplate:setup(combat)--,yoffset,zoffset)

	self.combat=combat
	self.character=combat.character
	
	self.resourcedisplays={}
	
	--self.displaywidth=160	--size of HP bar, including border
	self.displaywidth=self.character.modelinfo.clickwidth
	--TODO: dynamic nameplate height -- search for NAMEPLATE_HEIGHT and replace with something more useful
	self.displayheight=TRAININGGROUNDS_Nameplate.NAMEPLATE_HEIGHT
	
	-- hp shows up as red (on enemies)
	self.hpfraction=1
	-- damage shows up as some other undecided color
	self.hurtfraction=nil

	--TODO: nameplate:Cleanup()
	--TODO: remove yoffset/zoffset if we're sure we're not using them
	
	-- ""parent"" frame
	-- although the resource displays will SetPoint to this frame,
	-- they need to be parented to environment.frame or they won't respect scrollframe boundaries
	self.frame=TRAININGGROUNDS_ReusableFrames.getFrame()
	self.frame:SetWidth(100);self.frame:SetHeight(10)	--container frame size is arbitrary
	self.frame.parent=self.character.frame
	--self.frame.framelevel=self.frame.parent.framelevel+1
	self.frame.framelevel=255		--!!!
	self.frame:SetFrameLevel(self.frame.framelevel) --!!!
	--self.frame:SetPoint("CENTER",self.frame.parent,"CENTER",0,0)
	--self.frame:SetPoint("TOP",self.frame.parent,"TOP",0,0)
	self.frame:Show()
	--self.frame.texture:Show()
	--TRAININGGROUNDS_FrameVisibilityCheck(self.frame)
	
	
	self.yoffset=yoffset
	self.zoffset=zoffset

	local i
	for i=1,#self.combat.resources,1 do
		local resource=self.combat.resources[i]
		if(resource.visibleinnameplate==true)then			
			local resourcedisplay=resource:GetDisplay(self)
			resourcedisplay:setup(self,resource)
			tinsert(self.resourcedisplays,resourcedisplay)
		end
		--print("Nameplate.lua Added resource display")
		--print("Nameplate.lua Resource displays table: "..tostring(self.resourcedisplays))
		--print("Nameplate.lua Number of resource displays: "..tostring(#self.resourcedisplays))
	end
	
	local castbar=TRAININGGROUNDS_NameplateCastBar.new()
	castbar:setup(self,self.combat)
	self.castbar=castbar
	
	self:ResetPoint()

	--print("Nameplate.setup")
	
	--self.border.framelevel=self.character.frame.framelevel+1
	--self.border:SetFrameLevel(self.border.framelevel)
end

function TRAININGGROUNDS_Nameplate:SetAllFrameLevels(layer)
	for i=1,#self.resourcedisplays,1 do
		self.resourcedisplays[i]:SetFrameLevels(layer)
	end
end

function TRAININGGROUNDS_Nameplate:ResetPoint()
	local i
	local camera=self.character.environment.camera
	local scaleratio
	if(self.character.modelinfo.gameplayscaleratio~=nil)then
		scaleratio=self.character.modelinfo.gameplayscaleratio
	else
		scaleratio=1		
	end
	local extrascale
	if(self.character.extrascale~=nil)then
		extrascale=self.character.extrascale
	else
		extrascale=1		
	end
	local adjustedzoom=camera.zoom*scaleratio*extrascale
	
	local clickheight=self.character:GetClickareaHeightWithoutNameplate()
	local totalyoffset=(self.character.clickarea.y+clickheight)*adjustedzoom
	--local totalyoffset=(self.character.clickarea.height)*adjustedzoom
	self.frame:SetPoint("CENTER",self.frame.parent,"CENTER",0,totalyoffset)
	
	--TODO: need to go through list of resource bars
	-- and arrange them vertically
	--TODO: ask display what height it is
	
	local width=self.character.clickarea.width*adjustedzoom

	-- don't scale nameplates with character model scale (at the moment)
	--(TODO:)
	--local height=8*adjustedzoom	
	local height=8*camera.zoom
	for i=1,#self.resourcedisplays,1 do	
		local display=self.resourcedisplays[i]
		display:SetPointsAndSizes("TOP",self.frame,"TOP",0,(#self.resourcedisplays-i)*(height+2)+height,width,height)
	end
	self.castbar:SetPointsAndSizes("TOP",self.frame,"TOP",0,-1*(height+2)+height,width,height)
	
	
end

function TRAININGGROUNDS_Nameplate:step(elapsed)
	for i=1,#self.resourcedisplays,1 do
		local display=self.resourcedisplays[i]
		display:step(elapsed)
	end
	self.castbar:step(elapsed)
end

function TRAININGGROUNDS_Nameplate:GetNameplateHeight()
	-- nameplate height deliberately does not include cast bar,
	--  not that it makes a difference at the moment
	--TODO
	return 10*#self.resourcedisplays
end

-- this function currently expects to be called on hitpoints only
-- it also expects HP to be the topmost resource bar
function TRAININGGROUNDS_Nameplate:CreateDamageParticle(oldhp,newhp)
	local i
	local camera=self.character.environment.camera
	local scaleratio
	if(self.character.modelinfo.gameplayscaleratio~=nil)then
		scaleratio=self.character.modelinfo.gameplayscaleratio
	else
		scaleratio=1		
	end
	local extrascale
	if(self.character.extrascale~=nil)then
		extrascale=self.character.extrascale
	else
		extrascale=1
	end
	--local adjustedzoom=camera.zoom*scaleratio*extrascale
	local adjustedzoom=scaleratio*extrascale	
	
	local clickheight1=self.character:GetClickareaHeightWithoutNameplate() 
	
	local height=8--*adjustedzoom	
	local i=#self.resourcedisplays
	--local clickheight2=(#self.resourcedisplays-i)*(height+2)+height
	--TODO: hurtbox isn't lined up perfectly -- slightly too low -- cheap workaround -- fix it
	local clickheight2=(#self.resourcedisplays-i)*(height+2)+height+4
	
	--local totalyoffset=(self.character.clickarea.y+clickheight)*adjustedzoom
	local totalyoffset=(self.character.clickarea.y+clickheight1+clickheight2)*adjustedzoom
	--print(clickheight1,clickheight2)

	-- we should check for this before calling function
	if(newhp>oldhp)then return end
	
	--TODO: one function where nameplate can ask how wide it is (or how wide clickarea is)
	local barwidth=self.displaywidth*scaleratio*extrascale --*self.character.environment.camera.zoom
	local barheight=self.displayheight --*self.character.environment.camera.zoom
	
	local oldx=(oldhp/self.combat.hp.max)*barwidth	
	local newhpfraction=(newhp/self.combat.hp.max)
	local newx=newhpfraction*barwidth
	
	--local totalyoffset=(self.character.clickarea.y)
	--totalyoffset=totalyoffset+self.character.clickarea.height
	--totalyoffset=totalyoffset-TRAININGGROUNDS_Nameplate.NAMEPLATE_HEIGHT*.5
	
	local hurtbox=TRAININGGROUNDS_TemporaryParticleFrame.new()
	hurtbox:setup(self.character.environment)
	hurtbox.particlewidth=oldx-newx
	--hurtbox.x=self.character.x +  .5*(2+(barwidth-4)*newhpfraction)*adjustedzoom
	
	--local x=hurtbox.particlewidth/2
	--x=x+
	local x=(barwidth)*(-.5+newhpfraction)
	x=x+hurtbox.particlewidth/2
	
	hurtbox.x=self.character.x + x
	hurtbox.y=self.character.y
	local TEMP_screenscale=1	--1/1.5	--!!!
	hurtbox.z=(totalyoffset)/self.character.environment.camera.zstretch*TEMP_screenscale*TEMP_screenscale
	--print(totalyoffset)
	
	hurtbox.particleheight=(barheight-4)
	hurtbox:SetFrameSize()
	hurtbox:draw()	
	--hurtbox.frame.texture:SetTexture(0.75,0,0,1)
	hurtbox.frame.texture:SetTexture(1,0,0,1)
	hurtbox.frame.texture:Show()
	hurtbox.frame:Show()
	
	--hurtbox.xspeed=0
	--hurtbox.yspeed=0
	--hurtbox.zspeed=200
	
	
	
	-- if hurtbox is immobile, make sure it appears in front of resource bar
		-- (not 100% effective, and current hurtbox color is underwhelming)
	hurtbox.y=hurtbox.y-5
	--textbox.expirytime=self.currenttime+2000
	hurtbox.expirytime=self.character.environment.currenttime+1000
	
	return hurtbox
	
end

-- function TRAININGGROUNDS_Nameplate:SetNewHPRatio(fraction)
	-- --print("TRAININGGROUNDS_Nameplate:SetNewHPRatio")
	-- --TODO:
	-- self.hpfraction=fraction
	-- self:ResetPoint()
-- end

-- function TRAININGGROUNDS_Nameplate:SetNewHPRatioFromDamage(fraction)
-- local breakbar=false
-- if(self.hpfraction>0.75 and fraction<=0.75) then breakbar=true end --!!!
	-- if(fraction<self.hpfraction)then
		-- if(self.hurtfraction==nil)then 
			-- self.hurtfraction=0
		-- end		
		-- self.hurtfraction=self.hurtfraction+(self.hpfraction-fraction)
		-- --print("Nameplate:SetNewHPRatioFromDamage "..self.hurtfraction)
	-- end
	-- self:SetNewHPRatio(fraction)
	
-- --if(breakbar==true) then self:BreakHurtBar() end --!!!
-- end

-- -- this function returns the broken hurtbar so the calling function can toss it in a direction or break it further
-- function TRAININGGROUNDS_Nameplate:BreakHurtBar()
	-- --print("TRAININGGROUNDS_Nameplate:BreakHurtBar()")
	
	-- --TODO: recycled code; move to function
	-- local totalyoffset=(self.character.clickarea.y)
	-- totalyoffset=totalyoffset+self.character.clickarea.height
	-- totalyoffset=totalyoffset-TRAININGGROUNDS_Nameplate.NAMEPLATE_HEIGHT*.5
	-- local borderwidth=self.displaywidth --*self.character.environment.camera.zoom
	-- local borderheight=self.displayheight --*self.character.environment.camera.zoom
	
	-- local hurtbox=TRAININGGROUNDS_TemporaryParticleFrame.new()
	-- hurtbox:setup(self.character.environment)
	-- hurtbox.x=self.character.x +  .5*(2+(borderwidth-4)*self.hpfraction)
	-- hurtbox.y=self.character.y
	-- hurtbox.z=(totalyoffset)/self.character.environment.camera.zstretch
	-- --hurtbox.x=0;hurtbox.y=0 --!!!	
	-- hurtbox.particlewidth=(borderwidth-4)*self.hurtfraction
	-- hurtbox.particleheight=(borderheight-4)
	-- hurtbox:SetFrameSize()
	-- --hurtbox.frame:SetWidth((borderwidth-4)*self.hurtfraction)
	-- --hurtbox.frame:SetHeight(borderheight-4)
	-- hurtbox:draw()	
	-- hurtbox.frame.texture:SetTexture(0.75,0.75,0,1)
	-- hurtbox.frame.texture:Show()
	-- hurtbox.frame:Show()
	
	-- hurtbox.xspeed=0
	-- hurtbox.yspeed=0	
	-- --textbox.expirytime=self.currenttime+2000
	-- hurtbox.expirytime=self.character.environment.game.currenttime+4000
	
	-- self.hurtfraction=nil
	-- self:ResetPoint()
	
	-- return hurtbox
-- end


function TRAININGGROUNDS_Nameplate:Cleanup()
	for i=1,#self.resourcedisplays,1 do
		self.resourcedisplays[i]:Cleanup()
	end
	self.castbar:Cleanup()
	self.frame:Cleanup()
	--TODO: additional cleanup as necessary
end