TRAININGGROUNDS_Spell = {}
TRAININGGROUNDS_Spell.__index = TRAININGGROUNDS_Spell

function TRAININGGROUNDS_Spell:new()
	local self={}
	setmetatable(self, TRAININGGROUNDS_Spell)
		
	return self
end





function TRAININGGROUNDS_Spell:setup(combat)
	self.combat=combat
	self.button=nil		--this is set by SpellButton:setup
		
	self.offcooldowntick=0	-- time at which spell is no longer on cooldown 
	self.oncooldowntick=0	-- time at which spell went on cooldown (used to animate button cooldown swirls)
	self.naturaloffcooldowntick=0	-- time at which spell is no longer on cooldown, not counting GCDs	
	self.nextrechargetick=0	-- time at which spell gains a new charge
	
	
	
	self:setCustomSpellInfo(combat)
	if(self.defaultrecharge==nil)then
		self.defaultrecharge=self.defaultcooldown
	end

	self.charges=self.maxcharges
	
	self.target=nil		--if user changes target during cast, cast still applies to original target
	
	--casts are applied in CombatModule:TryExecuteSpell
	self.caststarttime=nil
	self.castendtime=nil
		
	
	--channels are applied in Spell:ExecuteSuccessfully
	-- combatmodule will keep track of the currently channeled spell
	-- the details of the channel are stored here
	self.channelstarttime=nil
	-- some of these might not matter until we stop snapshotting haste on channels
	self.channelendtime=nil	
	self.channelprevtick=nil	-- not currently used
	self.channelnexttick=nil	-- not currently used -- handled via combat "nextchannelticktime"

	
	self.usecount=0	-- mostly for statistics purposes
	self.usecounttextbox=nil	-- appears in debug_debugwindow
	
	
	
	
	--TODO: check if Floating Butterfly FoF animations exist somewhere
	
	--37: start jump
	--38: midair
	--39: land on ground
	--54: finish casting (use for chi wave)
	--61: mainhand drink
	--69: /dance
	--85,87,117,118(?): jab(?)
	--660: SCK
	--662: BoK
	--664: start rolling
	--666: rolling
	--668: stop rolling
	--670: tiger palm loop!?
	--672: monk unarmed attack mainhand 1
	--674: monk unarmed attack offhand 1
	--678: monk stance 1
	--680: chi burst
	--688: monk stance 2
	--682: monk unarmed attack mainhand 2
	--684: monk unarmed attack offhand 2
	--692: expel harm
	--698: monk stance 3
	--702: FSK
	--704: FSK start
	--706: FSK end, monk stance
	--712: FSK end, stance 0
	--716: fists of fury
	--726: soothing mist
	--728: unknown monk channel
	--730: chi torpedo
	--736: rising sun kick
end

function TRAININGGROUNDS_Spell:setCustomSpellInfo(combat)
	-- override as necessary	
	
	self.spellname="(ERROR: No spell name)"
	self.shortspellname="(ERROR)"
	self.usesgcd=true

	-- damage is automatically applied to target at impact time
	-- set damage to nil for AOE attacks
	self.damage=nil
	self.damagetype=TRAININGGROUNDS_Spell.TYPE_PHYSICAL
	
	--TODO: convert resource costs to list so we're compatible with rogues
	self.resource=nil
	self.resourcecost=0
	self.defaultcooldown=0
	self.maxcharges=1
	self.defaultrecharge=nil
	self.onlyshowrechargenevercooldown=false	--!!!  currently false to prevent cooldown draw bug, 
												-- but we should be able to set it to true without breaking everything
	--TODO: stop FoF/SCK animation when channel ends
	self.channeled=false
	self.channelduration=9999
	self.channeltotalticks=9
	-- tickonchannelstart counts as one of the channel total ticks
	self.tickonchannelstart=true
	
	--TODO: better name than "casted"
	self.casted=false
	self.casttime=nil
	--instant cast spells are generally castable while moving
	self.castablewhilemoving=true
	
	-- spells that require facing... don't, in this game
	-- but they will automatically turn the player towards the target for a moment
	self.requirefacing=false
	
	-- a few spells, such as Roll, aren't queueable
	self.queueable=true
	
	
	self.animationname=nil
	self.impacttime=nil
	
	--self.animationlength=9999
	self.animationsequence=1
	self.channeledanimationsequence=1
	
	

	
	--TODO: change default to false
	self.requirestarget=true
	self.range=50 -- melee range
	
	self.mainhanditemlength=9999
	self.mainhanditem=nil

	self.offhanditemlength=9999
	self.offhandhanditem=nil	
	
	self.icon="Interface\\Icons\\inv_misc_questionmark"
	
	self.procscombostrikes=false	--TODO: this should be in the 7.0 WW spell parent class
	
	self.visibletostatistics=true
end

function TRAININGGROUNDS_Spell:TargetIsOkay(character,target)
	if(self.requirestarget~=true)then
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)
	end	
	
	--override as necessary
	if(target==nil) then
		--TODO: attempt to target nearest enemy
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(false,"There is nothing to attack.")
	end
	
	--TODO: target being dead shouldn't necessarily disqualify the spell
	if(target.deadingame==true) then
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(false,"Invalid target")
	end
		
	return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)
end

-- This one checks for range only and doesn't return an error message.
--TODO: return error message
function TRAININGGROUNDS_Spell:RangeIsOkay(character,target)
	if(self.requirestarget~=true)then
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)	
	end
	if(self.range==nil)then
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)	
	end
	--TODO: take Z coordinate into account?
	local distance=character:HitboxDistanceFromOtherObject(target)
	if(distance>self.range)then	
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(false,"Out of range.")
	end
	return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)	
end

function TRAININGGROUNDS_Spell:haveResources()
	--override as necessary
	if(self.resource==nil)then
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)
	end	
			
	--print(tostring(self.resource.value) .. " " .. tostring(self.resourcecost))
	if(self.resource.value>=self.resourcecost)then
		return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(true,nil)
	end
	return TRAININGGROUNDS_SpellSuccessOrFailureResult_quicknew(false,"Not enough "..self.resource.name)
end

-- Call this function only if the spell is usable.
function TRAININGGROUNDS_Spell:executeSuccessfullyCastStart(executespelltick)	

	if(self.combat.combatstarttime==nil)then
		self.combat.combatstarttime=executespelltick
	end
	
	self:StopCastingOrChannelingOldSpell(executespelltick)

	--TODO: this breaks for pets that don't use caststart (e.g. SE&F)
	self.target=self.combat.target
	
	if(self.casted~=true)then
		self:executeSuccessfullyCastEnd(executespelltick)
	else
		self:ApplyCastIfApplicable(executespelltick)		
	end
	
	if(self.usesgcd==true)then
		self:ApplyGCD(executespelltick)		
		-- when cast finishes, we'll set combocount and previousgcdspells

	end
	
	--print(self,self.requirefacing)
	if(self.requirefacing==true)then
		--print("Set snapfacingtarget to ",self.combat.target)
		self.combat.character.snapfacingtarget=self.combat.target
		--TODO: not "500"
		self.combat.character.snapfacingendtick=executespelltick+500
		--print("End tick: ",self.combat.character.snapfacingendtick)
	end
	
	self.combat:DrawAllCooldowns(executespelltick)
end

-- Call this function only if spell is usable.
function TRAININGGROUNDS_Spell:executeSuccessfullyCastEnd(executespelltick)	
	--self.usecount=self.usecount+1
	self:IncrementSpellUseCount()
	--TODO: add up spell.usecount instead of using "totalspellusecount"
	self.combat.totalspellusecount=self.combat.totalspellusecount+1
	
	if(self.usecounttextbox~=nil)then
		self.usecounttextbox.frame.fontstring:SetText(self.shortspellname..": "..self.usecount)
	end
	
	--TODO: unless this spell can be used while casting/channeling...
	self:StopCastingOrChannelingOldSpell(executespelltick)

	--TODO: standardize checks between CreateSpellTextboxText, Combo Strikes damage multiplier / HitCombo
	--TODO: fix hardcoded player check
	if(self.combat.character:isa(TRAININGGROUNDS_PlayerCharacter))then
		local spelltext=self:CreateSpellTextboxText()
		self.combat.character.environment.game:createSpellTextbox(spelltext)		
	end
	
	if(self.usesgcd==true)then
		if(self.procscombostrikes==true)then
			if(self.combat.previousgcdspell~=self)then
				self.combat.lastattackwasokforcombostrikes=true
			else
				self.combat.lastattackwasokforcombostrikes=false
			end
		
			-- already applied GCD when cast started, but now we set combostrikes info
			-- when cast finishes, we'll set combocount and previousgcdspells
			self.combat:SetComboCount(executespelltick,self)
			self.combat.previous2gcdspell=self.combat.previousgcdspell
			self.combat.previousgcdspell=self		
		end
	end
	
	if(_TRAININGGROUNDS.game.targettalents==6)then
		if(self.usesgcd==true)then
			if(self.procscombostrikes==true)then
				if(self.combat.lastattackwasokforcombostrikes==true)then
					local hc=TRAININGGROUNDS_Buff_W7_WindwalkerMonk_HitCombo.new()
					hc:setup(self.combat,executespelltick)
				else
					local hc=self.combat:GetBuffIfExists(TRAININGGROUNDS_Buff_W7_WindwalkerMonk_HitCombo)
					if(hc~=nil)then
						hc:SelfDispel(executespelltick)
					end
				end
			end
		end	
	end
	
	self:ApplySpellAnimation(executespelltick)
	-- TODO: WARNING: we're currently assuming all spells are instant cast (or instant start-to-channel)
	self:applyExecutedSpellEffects(executespelltick)
	-- at the moment, FoF relies on spell effects being applied before channel is applied
	self:ApplyChannelIfApplicable(executespelltick)
	self:deductResourceCost()
	-- when we apply cooldown, we only want to redraw the cooldown if the CD actually changed.
	-- otherwise it looks weird.  so we need to remember the old cooldown
	local previousoffcooldowntick=self.offcooldowntick
	self:applyCooldown(executespelltick)
	self.combat.previousspell=self

	if(self.requirefacing==true)then
		self.combat.character.snapfacingtarget=self.combat.target
		--TODO: not "500"
		self.combat.character.snapfacingendtick=executespelltick+500
	end
	
	
	if(self.casted==true)then
		if(self.offcooldowntick~=previousoffcooldowntick)then
			self.combat:DrawOneCooldown(self,executespelltick)
		end
	else
		--if spell ISN'T casted, do nothing -- we just cast DrawAllCooldowns during startcast
	end
	
	--TODO: don't hardcode SEF check
	if(self.procscombostrikes==true)then		
		local i
		for i=1,#self.combat.pets,1 do			
			self.combat.pets[i].combat:TryMimicSpell(self,executespelltick)
		end
	end
		
end


function TRAININGGROUNDS_Spell:CreateSpellTextboxText()
	local text=self.spellname
	if(self.usesgcd==false)then
		text="|cffffffff("..text..")"
	else
		-- --text="|cff00ff00"..text.."!"
		-- text="|cff00ff00"..text..""
		
		-- Hit Combo
		if(self.procscombostrikes~=true)then
			text="|cff00ff00"..text
		else
			if(self~=self.combat.previousgcdspell)then
				text="|cff00ff00"..text.."!"
			else
				text="|cffffff00"..text
			end
		end
		-- -- T19 4pc		
		-- if(spell~=spell.combat.previousgcdspell
		-- and spell~=spell.combat.previous2gcdspell
		-- and spell.combat.previousgcdspell~=spell.combat.previous2gcdspell
		-- and spell~=nil and spell.combat.previousgcdspell~=nil and spell.combat.previous2gcdspell~=nil)then		
			-- text="|cff00ff00"..text.."!"
		-- else
			-- text="|cffffff00"..text
		-- end
	end
	return text
end


-- At the moment, only Flying Serpent Kick, SEF, and Xuen should use this function.
-- It doesn't check for resources, valid target, or distance to target.
function TRAININGGROUNDS_Spell:executeDespiteCooldown(executespelltick)
	--TODO: ApplySpellEffectsOnCooldown
	-- override
	
	-- if there was nothing to do, return false (will cause an error message to display later).
	-- otherwise return true
	return false
end

function TRAININGGROUNDS_Spell:StopCastingOrChannelingOldSpell(executespelltick)
	-- if new spell can be cast while casting/channeling old spell, override w/o calling super
	if(self.combat.castedspell~=nil)then
		self.combat:StopCasting()
	end
	if(self.combat.channeledspell~=nil) then
		--TODO:				
		self.combat:StopChanneling()
	end	
end

function TRAININGGROUNDS_Spell:applyExecutedSpellEffects(executespelltick)
	self:ApplyExecutedSpellDamage(executespelltick)
	self:ApplyExecutedSpellSideEffects(executespelltick)
end

function TRAININGGROUNDS_Spell:ApplyExecutedSpellDamage(executespelltick)
	--override when necessary (e.g. AOE attacks)
	if(self.impacttime~=nil)then
		local impacttick=executespelltick+self.impacttime
		self.combat:QueueImpact(impacttick, 
									function(tick)
										self:OnImpact(impacttick)
									end)
	else
		self:OnImpact(executespelltick)
	end		

end

function TRAININGGROUNDS_Spell:OnImpact(impacttick)
	if(self.damage~=nil and self.target~=nil)then
		self.combat:ApplyDamageToEnemy(self.damage,self.target,self)
	end
end

function TRAININGGROUNDS_Spell:ApplyExecutedSpellSideEffects(executespelltick)
	--override
end

-- casts are applied via combatmodule.TryExecuteSpell.
function TRAININGGROUNDS_Spell:ApplyCastIfApplicable(executespelltick)
	if(self.casted)then		
		self.combat:RevealNameplateCastbar()
		self.combat.castedspell=self
		self.caststarttime=executespelltick
		self.castendtime=executespelltick+self:GetCastTimeForUse()
		self:OnCastStart(self.caststarttime)
	end
end

-- channels are applied via ExecuteSuccessfully.
function TRAININGGROUNDS_Spell:ApplyChannelIfApplicable(executespelltick)
	--override and call super

	--TODO: move animation timer here? (currently in spell:executeSuccessfully)
	-- or at least have a separate channelanimation that plays after first animationtimer wears off

	if(self.channeled)then
		self.combat:RevealNameplateCastbar()
		self.combat.channeledspell=self
		self.channelstarttime=executespelltick
		local channeltime=floor(self.channelduration*self.combat:GetHasteCastReductionMultiplier())
		self.channelendtime=executespelltick+channeltime
		--print("Spell:ApplyChannelIfApplicable: "..self.channelendtime-self.channelstarttime)
		self:OnChannelStart(self.channelstarttime)
		
		--set player.channelanimation here
		
		self.combat.channelticksremaining=self.channeltotalticks
		
		if(self.tickonchannelstart==true)then			
			self:OnChannelTick(executespelltick)
			self.combat.channelticksremaining=self.combat.channelticksremaining-1
		end		
		
		self.combat.channeltickfrequency=channeltime/self.combat.channelticksremaining
		self.combat.nextchannelticktime=executespelltick+self.combat.channeltickfrequency
		
		
		--TODO: set nextidlegcdtime
	end	
end

function TRAININGGROUNDS_Spell:deductResourceCost()
	--override as necessary
	if(self.resource~=nil)then
		--TODO: instead of overriding deductResourceCost, implement CurrentResourceCost and override that instead
		self.resource:AddToValue(-self.resourcecost)
	end
end

function TRAININGGROUNDS_Spell:applyCooldown(executespelltick)
	if(_TRAININGGROUNDS.debug_nocooldowns~=true)then
		self.charges=self.charges-1
		--print(self.spellname," used at ",executespelltick)
		--print(self.spellname," has ",self.charges," charges remaining")
		if(executespelltick>=self.nextrechargetick)then
			if(self.nextrechargetick<executespelltick+self:GetRechargeForUse())then
				self.nextrechargetick=executespelltick+self:GetRechargeForUse()
			end			
		end				
		if(self.charges>=1)then
			--TODO: implement charges, which use GetCooldownForRecharge() instead
			if(self.offcooldowntick<executespelltick+self:GetCooldownForUse())then
				self.offcooldowntick=executespelltick+self:GetCooldownForUse()
			end
			self.naturaloffcooldowntick=self.offcooldowntick
			self.oncooldowntick=executespelltick
		else
			if(self.offcooldowntick<self.nextrechargetick)then	
				self.offcooldowntick=self.nextrechargetick
			end
			self.naturaloffcooldowntick=self.offcooldowntick			
			self.oncooldowntick=executespelltick			
		end
		--print(self.spellname," will come off cooldown at ",self.naturaloffcooldowntick)
	end
	self:OnApplyCooldown()
end

function TRAININGGROUNDS_Spell:ApplyGCD(executespelltick)
	--override as necessary
	self.combat:setGCD(executespelltick,1000)
end

-- this event triggers when the spell's cooldown is applied
-- via means other than GCD.  do not call this function on GCD.
function TRAININGGROUNDS_Spell:OnApplyCooldown()
	-- override as necessary
end

-- call this when the spell is used and there might be a reduce-cooldown-on-use effect
function TRAININGGROUNDS_Spell:GetCooldownForUse()
	--override as necessary
	return self.defaultcooldown
end

function TRAININGGROUNDS_Spell:GetRechargeForUse()
	--override as necessary
	--TODO: Serenity-affected spells need to override this in addition to overriding getcooldownforuse
	return self.defaultrecharge
end

-- call this when the spell recharges or whenever reduce-cooldown-on-use effect doesn't apply
function TRAININGGROUNDS_Spell:GetCooldownForRecharge()
	return self.defaultcooldown
end


function TRAININGGROUNDS_Spell:GetCastTimeForUse()
	return self.casttime*self.combat:GetHasteCastReductionMultiplier()
end


function TRAININGGROUNDS_Spell:OnCastStart(starttime)
	-- override
	--TODO: are we actually calling this?
end
--TODO: OnCastInterrupt
function TRAININGGROUNDS_Spell:OnCastEnd(endtime)
	-- override
	--TODO: are we actually calling this?
end

-- call this when casting stops for any reason, whether interrupted or reached the end of cast.
function TRAININGGROUNDS_Spell:OnCastStop(stoptime)
	--TODO: immediately return to idle animation ONLY IF channel animation wasn't overridden by another attack
	--we must be careful to always call this before setting the new attack's animation or comboresettime
	self.combat.character.attackanimationendtick=self.castendtime
	self.combat.character.combatstanceendtick=self.castendtime+3000
	self.combat.comboresettime=self.castendtime+500
end


function TRAININGGROUNDS_Spell:OnChannelStart(starttime)
	-- override
	-- some spells will also tick the moment they begin
	-- if so, we will call OnChannelTick from ApplyChannelIfApplicable
end

function TRAININGGROUNDS_Spell:OnChannelTick(ticktime)
	--TODO: set nextidlegcdtime
	-- channels that block other spells while channeling can probably set nextidlegcdtime to channel end time
	-- 		because the next queued spell will take effect
	-- channels that don't block other spells should give some leeway as casting exactly on time would override the last tick of the channel
end

function TRAININGGROUNDS_Spell:OnChannelStop(currenttime)	
	--TODO: immediately return to idle animation ONLY IF channel animation wasn't overridden by another attack
	--we must be careful to always call this before setting the new attack's animation or comboresettime
	self.combat.character.attackanimationendtick=self.channelendtime
	self.combat.character.combatstanceendtick=self.channelendtime+3000
	self.combat.comboresettime=self.channelendtime+500
end


function TRAININGGROUNDS_Spell:ApplySpellAnimation(executespelltime)
	if(self.animationname~=nil)then
		if(self.animationname=="FinishUndirectedSpellcast")then
			--TODO: don't hardcode exception
			--switch undirected spellcast to EMOTE after a moment so it doesn't interfere with running
			self.combat.character:SetAnimationByName(self.animationname,TRAININGGROUNDS_PlayerAnimations.PRIORITIES.EMOTE,500)
		else
			self.combat.character:SetAnimationByName(self.animationname)
		end
	end
	--if(self.animationsequence~=nil) then
		-- local totalanimationlength=self.animationlength
		-- --TODO: press esc to stop casting/channeling?
		-- if(self.channeled)then
			-- totalanimationlength=totalanimationlength*self.combat:GetHasteCastReductionMultiplier()
		-- end
		-- self.combat.character.attackanimationendtick=executespelltime+totalanimationlength
		-- self.combat.character:changeFrameAnimationAndOverride(self.animationsequence)
	-- end
end

function TRAININGGROUNDS_Spell:UpdateButtonChargeDisplay(currenttime)
	if(self.button~=nil)then
		if(self.maxcharges<=1)then
			self.button.charges:Hide()
		else
			self.button.charges:Show()
			self.button.charges.fontstring:SetText(self.charges)
		end
	end
end

--TODO: rename IncrementUseCount
function TRAININGGROUNDS_Spell:IncrementSpellUseCount()
	--override if necessary
	--(some spells like BoK override this to increment fakespells like CB:BoK)
	self.usecount=self.usecount+1
end

function TRAININGGROUNDS_Spell:Cleanup()
	-- TODO:
	
	-- (so far there's not much to do)
	-- (spellbutton is cleaned up via CombatModule:Cleanup)
end