-- TODO: should this be a GameObject?  (we doubt it)

TRAININGGROUNDS_Camera = {}
TRAININGGROUNDS_Camera.__index = TRAININGGROUNDS_Camera

function TRAININGGROUNDS_Camera.new()
	local self=setmetatable({}, TRAININGGROUNDS_Camera)	
	return self
end

function TRAININGGROUNDS_Camera:setup(environment)
	self.environment=environment
	
	self.x=0
	self.y=0
	
	self.ZOOM_CHANGE_SPEED=0.05
	-- self.MIN_ZOOM=0.5
	self.MIN_ZOOM=0.35
	
	self.zoom=1
	self.targetzoom=1
	-- modelangle determines the model's appearance, while ystretch determines the model's placement in the world.
	-- do not exceed -math.pi*(7.0/16).  at some point, SetFacing will flip backwards to try to keep the model "upright"
	self.modelangle=0

	-- don't set ystretch to exactly zero -- there are a few things that still divide by it
	self.ystretch=1
	self.zstretch=1
	self.ysort=false
	
	--TODO: move TOPLEFT to UI
	--self.objectanchor="BOTTOMLEFT"
	self.objectanchor="TOPLEFT"
	
	-- camera will follow this object
	self.camerafocus=nil
	
	-- some debug views use this to force models offcenter.
	-- it doesn't take ystretch into account, so side view offset must be very large
	self.permoffsety=0
	
	self.previouszoom=self.zoom
	self.zoomchanged=false
	
	self.forceupdate=false
	
	self.snaptofocus=true
	self.snapdistance=100	--gradually increases to force a snap
	self.catchupspeed=2
end

function TRAININGGROUNDS_Camera:step(elapsed)

	--TODO: properly reset each environment's camera upon return to main menu
	if(self.camerafocus~=nil)then
		if(self.camerafocus.dead==true)then
			self.camerafocus=nil
			self.zoom=1
			self.targetzoom=1
		end
	end

	if(self.zoom~=self.targetzoom)then
		local newzoom=self.zoom
		if(self.zoom>self.targetzoom)then
			newzoom=newzoom-1*(self.zoom-self.targetzoom)*elapsed
			if(newzoom<self.targetzoom)then
				newzoom=self.targetzoom
			end
		elseif(self.zoom<self.targetzoom)then
			newzoom=newzoom+1*(self.targetzoom-self.zoom)*elapsed
			if(newzoom>self.targetzoom)then
				newzoom=self.targetzoom
			end
		end
		self.zoom=newzoom
		--TODO:
		if(self.zoom<self.MIN_ZOOM)then self.zoom=self.MIN_ZOOM; self.targetzoom=self.zoom end
		if(self.zoom>1.5) then self.zoom=1.5; self.targetzoom=self.zoom end 	--!!!
	end

	if(self.zoom~=self.previouszoom or self.forceupdate==true) then
		self.zoomchanged=true
		self.forceupdate=false
	else
		self.zoomchanged=false
	end
	self.previouszoom=self.zoom
	
	self:MoveToScrollOffset(elapsed)
	
	local x,y=self:GetScrollOffset()
	self.environment.frame.scrollframe:SetHorizontalScroll(-x)
	self.environment.frame.scrollframe:SetVerticalScroll(y*self.ystretch)
	
end

function TRAININGGROUNDS_Camera:ChangeFocus(newfocus)
	self.camerafocus=newfocus
	self.snaptofocus=false
	self.snapdistance=100
	self.catchupspeed=2
end

function TRAININGGROUNDS_Camera:AdjustZoomViaScrollWheel(delta)
	self.targetzoom=self.targetzoom+delta*self.ZOOM_CHANGE_SPEED	
	--TODO:
	if(self.targetzoom<self.MIN_ZOOM-0.2) then self.targetzoom=self.MIN_ZOOM-0.2 end
	local maxzoom=1.5+.2
	--local maxzoom=6	
	if(self.targetzoom>maxzoom) then self.targetzoom=maxzoom end
	-- while it's possible to set target beyond min/max here,
	-- step event will automatically snap target to min/max when camera zoom reaches those limits
	
end

--TODO: rename LocalCoordsToYStretchCoords
-- also, check if we still use this function or not
function TRAININGGROUNDS_Camera:LocalCoordsToCameraCoords(x,y)
	if(true)then return x,y end --!!!
	--local xoffset,yoffset=self:GetScrollOffset()
	local xoffset,yoffset=0,0
	x=(x-xoffset)/self.zoom
	y=(y-yoffset)/self.zoom
	y=y/self.ystretch
	return x,y
end

function TRAININGGROUNDS_Camera:ConvertObjectCoordsToMouseoverCoords(x,y,obj)
	local envoffsetx,envoffsety
	local objoffsetx,objoffsety
	if(self.objectanchor=="CENTER") then
		envoffsetx=0;envoffsety=0
	end
	if(self.objectanchor=="TOPLEFT")then
		--TODO: don't hardcode screen size
		--return x+640/2,y+480/2
		envoffsetx=-640/2;envoffsety=480/2
	end
	
	if(obj.alignment=="CENTER")then
		objoffsetx=0;objoffsety=0
	end
	if(obj.alignment=="TOPLEFT")then
		objoffsetx=obj.frame:GetWidth()/2;objoffsety=-obj.frame:GetHeight()
		--objoffsetx=0;objoffsety=0
	end	
	
	return x+envoffsetx+objoffsetx,y+envoffsety+objoffsety
	
end

function TRAININGGROUNDS_Camera:MoveToScrollOffset(elapsed)
	local targetx,targety

	--print("Camerafocus: ",self.camerafocus)
	
	if(_TRAININGGROUNDS.debug_fixedcamera) then
		targetx,targety=0,0+self.permoffsety
	elseif(self.camerafocus==nil) then
		targetx,targety=0,0+self.permoffsety
	else
		
		-- -- "always centered"
		-- local xoffset,yoffset=-self.camerafocus.x*self.zoom,-self.camerafocus.y*self.zoom
		
		-- -- "shmup camera"
		if(self.environment.arenaradius==nil)then
			return 0,0	--TODO: return "always centered" instead?
		end
		--TODO: don't hardcode -50/+50
		local leftmostscroll=(-self.environment.arenaradius*self.zoom+TRAININGGROUNDS_Game.SCREENWIDTH/2)-50
		local rightmostscroll=(self.environment.arenaradius*self.zoom-TRAININGGROUNDS_Game.SCREENWIDTH/2)+50
		local hztotal=(-leftmostscroll+rightmostscroll)
		local ratio=(self.camerafocus.x+self.environment.arenaradius)/(self.environment.arenaradius*2)	
		local xoffset=-(leftmostscroll+hztotal*(ratio))
		-- local topmostscroll=(-self.environment.arenaradius*self.zoom*self.ystretch+TRAININGGROUNDS_Game.SCREENHEIGHT/2)-100
		-- local bottommostscroll=(self.environment.arenaradius*self.zoom*self.ystretch-TRAININGGROUNDS_Game.SCREENHEIGHT/2)+100
		local bottommostscroll=(-self.environment.arenaradius*self.zoom+TRAININGGROUNDS_Game.SCREENHEIGHT/2)---10
		local topmostscroll=(self.environment.arenaradius*self.zoom-TRAININGGROUNDS_Game.SCREENHEIGHT/2)+100--+10	
		local vttotal=(-bottommostscroll+topmostscroll)
		local ratio=(self.camerafocus.y+self.environment.arenaradius)/(self.environment.arenaradius*2)
		local yoffset=-(bottommostscroll+vttotal*(ratio))
		
		targetx,targety=xoffset,yoffset+self.permoffsety
		--print("Camera:getscrolloffset ratio: "..ratio.." vttotal: "..vttotal)
	end
	
		
	--TODO: this could break spectacularly at low framerates
	-- 	...and it's probably not framerate-independent either.

	--self.snapdistance=self.snapdistance+elapsed*100
	--self.catchupspeed=self.catchupspeed+elapsed*1
	self.catchupspeed=self.catchupspeed+elapsed*2

	if(self.snaptofocus~=true)then
		self.x=self.x+(targetx-self.x)*self.catchupspeed*elapsed
		self.y=self.y+(targety-self.y)*self.catchupspeed*elapsed	
		--TODO: proper sqrt distance?
		if(math.abs(targetx-self.x)<self.snapdistance*elapsed and math.abs(targety-self.y)<self.snapdistance*elapsed)then self.snaptofocus=true end
	end
	
	if(self.snaptofocus==true)then
		self.x=targetx
		self.y=targety
	end
end

function TRAININGGROUNDS_Camera:GetScrollOffset()
	return self.x,self.y

end

function TRAININGGROUNDS_Camera:Cleanup()
	--TODO: cleanup
end