
TRAININGGROUNDS_Environment = {}
TRAININGGROUNDS_Environment.__index = TRAININGGROUNDS_Environment

function TRAININGGROUNDS_Environment:new()
	local self={}
	setmetatable(self, TRAININGGROUNDS_Environment)	

	return self
end

function TRAININGGROUNDS_Environment:setup(game,gameframe)	
--print("!!!",game,gameframe)
	self.game=game
	
	--self.frame=game.gameframe	-- TODO: individual frames for each environment?
	self.frame=TRAININGGROUNDS_ReusableScrollFrames.getFrame()
	self.frame:SetWidth(640)	--TODO: 
	self.frame:SetHeight(480)
	self.frame.parent=gameframe
	self.frame:SetParent(self.frame.parent)
	self.frame.scrollframe:SetParent(self.frame.parent)
	self.frame.scrollframe:SetWidth(640)	--TODO: 
	self.frame.scrollframe:SetHeight(480)
	self.frame:SetAllPoints(gameframe)
	self.frame.scrollframe:SetAllPoints(gameframe)
	self.frame:Show()
	self.frame.scrollframe:Show()
	self.frame.scrollframe:SetScrollChild(self.frame)
	
	self:ResetCamera()

	self.arenaradius=nil
	self.arenaborderwidth=nil
	
	self.newgameobjects={}
	self.gameobjects={}
	self.newwowcharacters={}
	self.wowcharacters={}	
	
	self.shadowframelevel=98 --self.ysortframelevel-2
	self.targetcircleframelevel=99 --self.ysortframelevel-1	
	-- if ysort is turned on, sorted objects will be drawn starting at this layer
	self.ysortframelevel=100
	
	
	self.nextmousehighlighttargetscantime=0			-- mouseover detection doesn't run every frame
	self.nextmousehighlighttargetscaninterval=100	
	self.nextysorttime=0			-- ysort doesn't run every frame
	--self.nextysortinterval=200		
	self.nextysortinterval=32		-- ...or does it
	
	self.currenttime=0
	
	tinsert(game.environments,self)
end

function TRAININGGROUNDS_Environment:step(elapsed)
	
	self.currenttime=self.currenttime+elapsed*1000

	for i=1,#self.newgameobjects,1 do
		tinsert(self.gameobjects,self.newgameobjects[i])				
	end
	self.newgameobjects={}
	for i=1,#self.newwowcharacters,1 do
		tinsert(self.wowcharacters,self.newwowcharacters[i])
	end
	self.newwowcharacters={}	
	
	
	
	local uiscale,globalx,globaly=UIParent:GetEffectiveScale(),GetCursorPosition()	
	local gamescale=self.game.frame:GetScale()
	-- must divide by uiscale
	-- TODO: do we also need to divide by gamescale?  environment_world divides by gamescale
	-- (for now we are dividing by gamescale)
	globalx=globalx/uiscale/gamescale
	globaly=globaly/uiscale/gamescale
	local left,bottom,width,height=self.frame:GetRect()
	--TODO: revise math and logic
		
	local localx=globalx-left
	local localy=globaly-bottom			
	--local offsetx,offsety=self.camera:GetScrollOffset()	
	local offsetx,offsety=0,0
	offsety=offsety*self.camera.ystretch
	offsetx=offsetx*self.camera.zoom
	offsety=offsety*self.camera.zoom
	offsetx=offsetx+TRAININGGROUNDS_Game.SCREENWIDTH/2
	offsety=offsety+(TRAININGGROUNDS_Game.SCREENHEIGHT/2)
	localx=localx-offsetx
	localy=localy-offsety
	local camerax,cameray=localx,localy
	camerax=camerax*(1/self.camera.zoom)
	cameray=cameray*(1/self.camera.zoom)
	cameray=cameray*(1/self.camera.ystretch)	
	--print("Environment.lua cursor "..camerax..","..cameray)
	
	------------- Mouseover target highlight ------------
	--TODO: sort by frame level
--print("self.game: ",self.game)	
	if(self.game.realtime>self.nextmousehighlighttargetscantime)then
		self.nextmousehighlighttargetscantime=self.game.realtime+self.nextmousehighlighttargetscaninterval
		for i=1,#self.gameobjects,1 do
			local obj=self.gameobjects[i]
			if(obj.clickable==true)then				
				obj.ismouseovered=false
				-- since we're operating in camera coordinates, divide by ystretch
				-- TODO: overridable GetClickareaRect
				local adjustedcamerax,adjustedcameray=localx,localy
				local modelratio=1
				if(obj.modelinfo~=nil)then					
					modelratio=obj.modelinfo.gameplayscaleratio*obj.extrascale
					--print("New modelratio is "..modelratio)
				end
				local adjustedzoom=self.camera.zoom*modelratio
				adjustedcamerax=adjustedcamerax*(1/adjustedzoom)
				adjustedcameray=adjustedcameray*(1/adjustedzoom)
				adjustedcameray=adjustedcameray*(1/self.camera.ystretch)	
				
				local x1=obj.x+(obj.clickarea.x-obj.clickarea.width/2)
				local x2=obj.x+(obj.clickarea.x+obj.clickarea.width/2)
				local y1=obj.y+(obj.clickarea.y)/self.camera.ystretch
				local y2=obj.y+(obj.clickarea.y+obj.clickarea.height)/self.camera.ystretch
				x1,y1=self.camera:ConvertObjectCoordsToMouseoverCoords(x1,y1,obj)
				x2,y2=self.camera:ConvertObjectCoordsToMouseoverCoords(x2,y2,obj)
				--print("Environment.lua "..x1..","..y1.." to "..x2..","..y2)
				if(adjustedcamerax>=x1)then
					if(adjustedcamerax<=x2)then						
						if(adjustedcameray>=y1)then
							if(adjustedcameray<=y2)then	
								--print("Environment.lua mouseover "..tostring(obj))
								obj.ismouseovered=true
							end
						end
					end
				end
			end
		end			
	end
	
	
	

	
	
	for i=1,#self.gameobjects,1 do
		self.gameobjects[i]:step(elapsed,self.currenttime)
	end
	
	-- sorting the table doesn't appear to cause a framerate drop, but actually changing framelevels does
	-- TODO: is it actually caused by changing framelevels, or just by having a lot of objects on the screen at once?
	
	self.camera:step(elapsed)	
	
	if(self.camera.ysort==true) then
		if(self.game.realtime>self.nextysorttime)then
			self.nextysorttime=self.game.realtime+self.nextysortinterval
			local tempysort={}
			for i=1,#self.gameobjects,1 do 
				if(self.gameobjects[i].ysortable==true)then
					--print("Game.lua: "..tostring(v).. v.y)
					tinsert(tempysort,self.gameobjects[i]) 
				end
			end
			table.sort(tempysort, function(a,b) return a.y>b.y end)
			local layer=self.ysortframelevel
			for k,v in ipairs(tempysort) do 	
				--TODO: move to overrideable function that takes level number as argument and returns number of used level
				v.tempframelayer=layer			
				v.frame.framelevel=layer
				v.frame:SetFrameLevel(layer)
				if(v.modelframe~=nil)then 
					v.modelframe.framelevel=layer
					v.modelframe:SetFrameLevel(layer) 					
				end	
				if(v.nameplate~=nil)then
					layer=layer+1
					v.nameplate:SetAllFrameLevels(layer)
				end
				-- if(self.player~=nil and v==self.player)then
					--print("Environment.lua: Player is now on layer "..layer)
				-- end
				layer=layer+1
			end
		end
	end

	
	for i=1,#self.gameobjects,1 do
		self.gameobjects[i]:animate()
	end	
	
	for i=1,#self.gameobjects,1 do
		self.gameobjects[i]:draw()			
	end


	
	
	
	--print("Game.lua: "..tostring(self.player).. "="..self.player.tempframelayer.." , " ..tostring(self.npc).."="..self.npc.tempframelayer)
	
	
	
		
	for i=#self.gameobjects,1,-1 do
		local obj=self.gameobjects[i]
		if(obj.dying==true) then
			obj:Cleanup()	--this SHOULD set dead to true
			tremove(self.gameobjects,i)
			--TODO: remove from wowobject list (maybe in wowobject:cleanup?)
		end
	end
	
end


-- returns a list of all gameobject within range of the line segment
function TRAININGGROUNDS_Environment:LineCircleCollisionList(x1, y1, x2, y2, linewidth)
	local list={}
	for i=1,#self.gameobjects,1 do
		local obj=self.gameobjects[i]		
		if(TRAININGGROUNDS_LineCircleCollision(x1,y1,x2,y2,obj.x,obj.y,linewidth+obj:HitboxRadius())==true)then
			tinsert(list,obj)
		end
	end
	return list
end

-- returns a list of all gameobject within range of the collision pie
-- anglesweep is +/- from angle, so total arc will be 2x anglesweep
-- gameobjects are treated as points instead of circles
--TODO: use LineCircleCollisionList to pick up circular objects
function TRAININGGROUNDS_Environment:PointPieCollisionList(x, y, angle, anglesweep, minradius, maxradius)
	local angle1=angle-anglesweep
	local angle2=angle+anglesweep	
	local list={}
	for i=1,#self.gameobjects,1 do
		local obj=self.gameobjects[i]
		local distsqr=(obj.x-x)*(obj.x-x)+(obj.y-y)*(obj.y-y)
		if(distsqr>=minradius*minradius  and distsqr<=maxradius*maxradius)then
			local objangle=math.atan2(obj.y-y,obj.x-x)
			if(angle1>objangle)then
				angle1=angle1-math.pi*2
				angle2=angle2-math.pi*2
			end
			if(angle2<objangle)then
				angle1=angle1+math.pi*2
				angle2=angle2+math.pi*2				
			end
			if(angle1<=objangle and objangle<=angle2)then
				tinsert(list,obj)
			end
		end
	end
	return list
end
--TODO: account for gameobject's radius
function TRAININGGROUNDS_Environment:PointCircleCollisionList(x, y, minradius, maxradius)
	local list={}
	for i=1,#self.gameobjects,1 do
		local obj=self.gameobjects[i]
		local distsqr=(obj.x-x)*(obj.x-x)+(obj.y-y)*(obj.y-y)
		if(distsqr>=minradius*minradius  and distsqr<=maxradius*maxradius)then
			tinsert(list,obj)
		end
	end
	return list
end






function TRAININGGROUNDS_Environment:ClearAllObjects()
	-- Clears all objects and resets the camera,
	-- but doesn't clean up anything else.
		
	for i=1,#self.gameobjects,1 do
		local obj=self.gameobjects[i]
		obj:Cleanup()		
	end
	self.gameobjects={}
	
	self:ResetCamera()
end

function TRAININGGROUNDS_Environment:ResetCamera()
	if(self.camera~=nil)then self.camera:Cleanup() end
	self.camera=TRAININGGROUNDS_Camera.new()
	self.camera:setup(self)
end

function TRAININGGROUNDS_Environment:Cleanup()
	--print("Environment contains ",#self.gameobjects," objects")
	for i=1,#self.gameobjects,1 do
		--print("Environment cleaning up ",obj)
		local obj=self.gameobjects[i]
		obj:Cleanup()		
	end
	self.gameobjects={}
	
	self.camera:Cleanup()
	
	self.frame:Cleanup()
	
	--TODO:
end




