TRAININGGROUNDS_Game = {}
TRAININGGROUNDS_Game.__index = TRAININGGROUNDS_Game

function TRAININGGROUNDS_Game.new()
	local self=setmetatable({}, TRAININGGROUNDS_Game)	
	return self
end

-- SCREENWIDTH and SCREENHEIGHT are the dimensions of the main game frame.
TRAININGGROUNDS_Game.SCREENWIDTH=640
--TRAININGGROUNDS_Game.SCREENWIDTH=1200	-- for debug menus
TRAININGGROUNDS_Game.SCREENHEIGHT=480
-- The actual window itself is slightly larger and contains the outer border and title bar.
TRAININGGROUNDS_Game.BORDERSIZE=2
TRAININGGROUNDS_Game.TITLEHEIGHT=20


function TRAININGGROUNDS_Game:setup()
	self.realtime=0
	self.currenttime=0
	self.keys=TRAININGGROUNDS_Keys.new()
	
	self.transitionfunction=nil
	self.transitionprogress=0	
	self.transitiondone=false

	TRAININGGROUNDS_PlayerAnimations.setup()
	
	self.alreadyseenmenuintro=false
	
	-- even though we store environments in a list, we still need to keep track of them in dedicated variables
	self.environments={}
	
	--TODO: set framelevel?
	--WARNING: this isn't a temp frame -- don't try to overwrite it
	if(self.frame==nil) then
		
		self.frame=CreateFrame("Frame",nil,UIParent)
		self.frame:SetParent(UIParent)		
		self.frame:SetWidth(TRAININGGROUNDS_Game.SCREENWIDTH+TRAININGGROUNDS_Game.BORDERSIZE*2)
		self.frame:SetHeight(TRAININGGROUNDS_Game.SCREENHEIGHT+TRAININGGROUNDS_Game.TITLEHEIGHT+TRAININGGROUNDS_Game.BORDERSIZE*2)
		local debugxoffset=0
		--if(_TRAININGGROUNDS.debug_debugwindow==true)then
		if(true)then
			debugxoffset=-100
		end
		--self.frame:SetPoint("CENTER",0+debugxoffset,0)
		self.frame:SetPoint("CENTER",0,0)
		if(not (_TRAININGGROUNDS.debug_dontscreenclamp==true))then
			self.frame:SetClampedToScreen(true)
		end
		self.frame.framelevel=1
		self.frame:SetFrameLevel(self.frame.framelevel)
		self.frame.texture=self.frame:CreateTexture()
		self.frame.texture:SetAllPoints(self.frame)
		self.frame.texture:SetTexture(0,0,0,0.75)
		--self.frame.texture:SetTexture(0.5,0.5,0.5,0.25)
		--self.frame.texture:SetTexture(1,1,1,0.0625)		
		self.frame.texture:Show()
		self.frame:Show()
		self:SetGameWindowScale()
		--print("Game:setup width: "..GetScreenWidth()..", height: "..GetScreenHeight()..", uiscale: "..UIParent:GetEffectiveScale())
		--print("Game:setup resolution: "..({GetScreenResolutions()})[GetCurrentResolution()])
		
		local title=self.frame:CreateTitleRegion()				
		title:SetPoint("TOPLEFT",self.frame,"TOPLEFT",0,0)		-- remember, this is the draggable title bar we're setting,
		title:SetPoint("TOPRIGHT",self.frame,"TOPRIGHT",0,0)	-- so use TOPLEFT/TOPRIGHT instead of BOTTOMLEFT/BOTTOMRIGHT
		title:SetHeight(TRAININGGROUNDS_Game.TITLEHEIGHT)
		-- we need to enable mouse for frame instead of gameframe to allow draggable title region,
		-- and we might as well enable mousewheel for frame instead just to keep everything together
		self.frame:EnableMouse(true)
		self.frame:EnableMouseWheel(true)
		-- we can't disable mouse without breaking draggable title region,
		-- so we keep track of game mouse with another variable
		self.mouse_input_enabled=true
		
		
		self.gameframe=CreateFrame("Frame",nil,UIParent)
		self.gameframe:SetParent(self.frame)		
		self.gameframe:SetWidth(TRAININGGROUNDS_Game.SCREENWIDTH)
		self.gameframe:SetHeight(TRAININGGROUNDS_Game.SCREENHEIGHT)
		self.gameframe:SetPoint("BOTTOM",self.frame,"BOTTOM",0,TRAININGGROUNDS_Game.BORDERSIZE)
		self.gameframe.framelevel=1
		self.gameframe:SetFrameLevel(self.gameframe.framelevel)
		-- -- for now, use self.frame's texture instead
		-- self.gameframe.texture=self.frame:CreateTexture()
		-- self.gameframe.texture:SetAllPoints(self.frame)
		-- self.gameframe.texture:SetTexture(0,0,0,0.75)
		-- self.gameframe.texture:Show()						
		self.gameframe:Show()		
		
		-- self.gamescroll=CreateFrame("ScrollFrame",nil,UIParent)
		-- self.gamescroll:SetParent(self.frame)		
		-- self.gamescroll:SetWidth(TRAININGGROUNDS_Game.SCREENWIDTH)
		-- self.gamescroll:SetHeight(TRAININGGROUNDS_Game.SCREENHEIGHT)
		-- self.gamescroll:SetPoint("BOTTOM",self.frame,"BOTTOM",0,TRAININGGROUNDS_Game.BORDERSIZE)
		-- self.gamescroll:SetScrollChild(self.gameframe)
		-- self.gamescroll.framelevel=1
		-- self.gamescroll:SetFrameLevel(self.gamescroll.framelevel)
		-- self.gamescroll:Show()
		
		self.border=TRAININGGROUNDS_BorderGraphicWithTitle.new()
		--!!!
		self.border:setup(self.frame,_TRAININGGROUNDS.WINDOWCAPTION,TRAININGGROUNDS_Game.TITLEHEIGHT)		
		--self.border:setup(self.frame,"TrainingGrounds - 2016-03-18",TRAININGGROUNDS_Game.TITLEHEIGHT)
		--self.border:setup(self.frame,"Work in progress",TRAININGGROUNDS_Game.TITLEHEIGHT)

		
		

		
		
		
		self.closebutton=TRAININGGROUNDS_ReusableButtonFrames.getFrame()
		self.closebutton:SetParent(self.frame)
		self.closebutton:SetWidth(16)
		self.closebutton:SetHeight(16)
		self.closebutton:SetPoint("TOPRIGHT",-5,-2)
		self.closebutton:SetFrameLevel(255)
		self.closebutton.texture=self.closebutton:CreateTexture()
		self.closebutton.texture:SetAllPoints(self.closebutton)		
		self.closebutton.texture:SetTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\closebutton.tga")	
		self.closebutton:SetHighlightTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\blanksquare.tga")	
		--self.closebutton:SetHighlightTexture("Interface\\AddOns\\TrainingGrounds\\Assets\\Graphics\\blanksquarehalfalpha.tga")	
		self.closebutton.texture:Show()		
		self.closebutton:Show()		
		self.closebutton:SetScript("OnClick",function(self,button,down)
				_TRAININGGROUNDS:Shutdown()
			end)

		
		
		--if(_TRAININGGROUNDS.debug_debugwindow==true)then
		if(true)then
			self.debugwindow=TRAININGGROUNDS_DebugWindow.new()
			
			self.debugwindow:setup(self)
			self.debugwindow.frame:Hide()
			--print("Debugwindow.game: ",self,self.debugwindow.game)
		end
		
		
		self.transitionfader=TRAININGGROUNDS_ReusableFrames.getFrame()
		self.transitionfader:SetWidth(TRAININGGROUNDS_Game.SCREENWIDTH);self.transitionfader:SetHeight(TRAININGGROUNDS_Game.SCREENHEIGHT)
		self.transitionfader:SetParent(self.gameframe)
		self.transitionfader:SetAllPoints(self.gameframe)
		self.transitionfader:SetFrameLevel(512)
		self.transitionfader.texture:SetTexture(0,0,0,0)
		self.transitionfader.texture:Show()
		self.transitionfader:Show()
		
		
	else
		--else restore window
	end
	
	
	-- -------------- -!!!
	-- local cd
	-- cd=CreateFrame("Cooldown",nil,PlayerFrame,"CooldownFrameTemplate")
	-- cd:SetAllPoints(PlayerPortrait)
	-- cd:Show()
	-- cd:SetCooldown(GetTime(),10)
	
	-- -------------------

	if(_TRAININGGROUNDS.debug_alwaysresetwarning==true)then
		TRAININGGROUNDS_VARS.acceptedwarning=false
	end

	


	if(TRAININGGROUNDS_VARS.acceptedwarning~=true and _TRAININGGROUNDS.debug_autoacceptwarning~=true)then
		self.environment_warningmenu=TRAININGGROUNDS_Environment_WarningMenu.new()
		self.environment_warningmenu:setup(self,self.gameframe)
	else
		self:ShowFirstMenuAfterWarning(true)
	end

	
	--print("Game.lua: "..tostring(self.player).." "..tostring(self.npc).." "..tostring(self.wowcharacters[1]).." "..tostring(self.wowcharacters[2]))
	
end

function TRAININGGROUNDS_Game:ShowFirstMenuAfterWarning(skippedwarning)
	self.environments={}
	
	if(self.debugwindow~=nil)then
		self.debugwindow.environment_ui=TRAININGGROUNDS_Environment.new()
		self.debugwindow.environment_ui:setup(self,self.debugwindow.gameframe)
	end
	
	self.environment_menu=TRAININGGROUNDS_Environment_Menu.new()		
	self.environment_menu:setup(self,self.gameframe)		
	self.environment_menu.frame:Show()
	--self.environment_menu.frame.framelevel=129
	--self.environment_menu.frame.scrollframe:SetFrameLevel(self.environment_menu.frame.framelevel)
	-- -- setting menu framelevel to 255 causes fakePCs to show up in front of menubuttons
	-- self.environment_menu.frame.framelevel=255
	-- self.environment_menu.frame.scrollframe:SetFrameLevel(self.environment_menu.frame.framelevel)
	--
	
	--TODO: this might be why the game softlocks at title screen if automenu is on
	if(_TRAININGGROUNDS.debug_automenu~=true)then
		local menu_titlescreen=TRAININGGROUNDS_MenuController_TitleScreen.new()
		menu_titlescreen:setup(self.environment_menu)
		if(skippedwarning==true)then	
			menu_titlescreen.logo.fulltitleprogress=48-16
		end
		if(self.alreadyseenmenuintro==true)then
			menu_titlescreen:ChangeMenuMode(2)
		end
	end
	
	--local menu_chooseclass=TRAININGGROUNDS_MenuController_ChooseClass.new()
	--menu_chooseclass:setup(self.environment_menu)
	-- local menu_alignclass=TRAININGGROUNDS_MenuController_ClassModelAlignment.new()
	-- menu_alignclass:setup(self.environment_menu)		
			
	
	self.environment_ui=TRAININGGROUNDS_Environment_UI.new()
	self.environment_ui:setup(self,self.gameframe)
	--self.environment_ui.frame.framelevel=255
	self.environment_ui.frame.framelevel=129
	--print("UI: ",self.environment_ui.frame,self.environment_ui.frame.framelevel)
	self.environment_ui.frame.scrollframe:SetFrameLevel(self.environment_ui.frame.framelevel)
	self.environment_ui.frame:Hide()
	
	self.environment_world=TRAININGGROUNDS_Environment_World.new()
	self.environment_world:setup(self,self.gameframe)	
	self.environment_world.frame.framelevel=4
	self.environment_world.frame.scrollframe:SetFrameLevel(self.environment_world.frame.framelevel)
	self.environment_world.frame:Hide()


	if(_TRAININGGROUNDS.debug_automenu==true)then
		--print("Automenu")
		self:Automenu()
	end
end


function TRAININGGROUNDS_Game:StartGame()
	--these values are set by menucontroller_titlescreen
	
	if(self.targetstage==-1)then
		self:SetEnvironmentTransition(function()
												self.environment_ui.frame:Show()
												self.environment_world.frame:Show()
												--self.environment_menu.frame:Show()												
												
												local scenario_tutorial=TRAININGGROUNDS_ScenarioController_Sandbox.new()
												scenario_tutorial:setup(self.environment_world)
												local pausemenu=TRAININGGROUNDS_MenuController_PauseMenu.new()
												pausemenu:setup(self.environment_ui)
											end)
	elseif(self.targetstage==1)then
		self:SetEnvironmentTransition(function()
												self.environment_ui.frame:Show()
												self.environment_world.frame:Show()
												--self.environment_menu.frame:Show()
												
												local scenario_tutorial=TRAININGGROUNDS_ScenarioController_Tutorial.new()
												scenario_tutorial:setup(self.environment_world)
												local pausemenu=TRAININGGROUNDS_MenuController_PauseMenu.new()
												pausemenu:setup(self.environment_ui)												
											end)
	elseif(self.targetstage==2)then
		self:SetEnvironmentTransition(function()
												self.debugwindow.frame:Show();self.debugwindow:RepositionDebugWindowForTheFirstTime()
												self.environment_ui.frame:Show()
												self.environment_world.frame:Show()
												--self.environment_menu.frame:Show()
												
												local scenario_tutorial=TRAININGGROUNDS_ScenarioController_EndlessWaves.new()
												scenario_tutorial:setup(self.environment_world)
												local pausemenu=TRAININGGROUNDS_MenuController_PauseMenu.new()
												pausemenu:setup(self.environment_ui)	--!!!
												
											end)	
	elseif(self.targetstage==3)then
		self:SetEnvironmentTransition(function()
												self.debugwindow.frame:Show();self.debugwindow:RepositionDebugWindowForTheFirstTime()
												self.environment_ui.frame:Show()
												self.environment_world.frame:Show()
												--self.environment_menu.frame:Show()
												
												local scenario_tutorial=TRAININGGROUNDS_ScenarioController_SustainedWorkout.new()
												scenario_tutorial:setup(self.environment_world)
												local pausemenu=TRAININGGROUNDS_MenuController_PauseMenu.new()
												pausemenu:setup(self.environment_ui)												
											end)	
	elseif(self.targetstage==4)then
		self:SetEnvironmentTransition(function()
												self.debugwindow.frame:Show();self.debugwindow:RepositionDebugWindowForTheFirstTime()
												self.environment_ui.frame:Show()
												self.environment_world.frame:Show()
												--self.environment_menu.frame:Show()
												
												local scenario_tutorial=TRAININGGROUNDS_ScenarioController_SustainedWorkout.new()
												scenario_tutorial:setup(self.environment_world)
												local pausemenu=TRAININGGROUNDS_MenuController_PauseMenu.new()
												pausemenu:setup(self.environment_ui)												
											end)
											
	else
		print("Something went wrong... going to tutorial.")
		self:SetEnvironmentTransition(function()
												self.environment_ui.frame:Show()
												self.environment_world.frame:Show()
												--self.environment_menu.frame:Show()
												
												local scenario_tutorial=TRAININGGROUNDS_ScenarioController_Tutorial.new()
												scenario_tutorial:setup(self.environment_world)
												local pausemenu=TRAININGGROUNDS_MenuController_PauseMenu.new()
												pausemenu:setup(self.environment_ui)												
											end)
	end	
end


function TRAININGGROUNDS_Game:Automenu()
	self.targetstage=1
	--self.targetstage=2
	--self.targetstage=3
	--self.targetstage=-1
	self.targettalents=1
	self.targethaste=1
	self:StartGame()
end

function TRAININGGROUNDS_Game:QuitLevelAndReturnToMenu()
	self.alreadyseenmenuintro=true
	self:SetEnvironmentTransition(function()
						self:CleanupGameplayEnvironmentsAfterReturningToMenu()
					end)
end

function TRAININGGROUNDS_Game:SetEnvironmentTransition(fn)
	if(self.transitionfunction~=nil)then return end
	
	self.transitionfunction=fn
	self.transitionprogress=0
	self.transitiondone=false	
end

function TRAININGGROUNDS_Game:SetGameWindowScale()
	if(_TRAININGGROUNDS.debug_dontautosize==true)then
		self.frame:SetScale(1)
		if(self.debugwindow~=nil)then	self.debugwindow.frame:SetScale(1)	end
	else
		self.frame:SetScale(1/UIParent:GetEffectiveScale())				
		if(self.debugwindow~=nil)then	self.debugwindow.frame:SetScale(1/UIParent:GetEffectiveScale())	end
		--self.frame:SetScale(1/UIParent:GetEffectiveScale()*1.5)
	end
	if(self.environments~=nil)then
		for i,v in ipairs(self.environments) do
			if(v.camera~=nil)then
				v.camera.forceupdate=true
			end
		end
	end
end

function TRAININGGROUNDS_Game:OnKeyChange(key,state)
	--print("TRAININGGROUNDS_Game:OnKeyChange "..key)
	
	-- TODO: convert to loop
	if(key=="UP" or key=="W")then		
		self.keys.up:SetKey(state)
	elseif(key=="DOWN" or key=="S")then
		self.keys.down:SetKey(state)
	elseif(key=="LEFT" or key=="A")then
		self.keys.left:SetKey(state)
	elseif(key=="RIGHT" or key=="D")then
		self.keys.right:SetKey(state)
	elseif(key=="Z")then
		self.keys.z:SetKey(state)
	elseif(key=="X")then
		self.keys.x:SetKey(state)
	elseif(key=="SPACE")then
		self.keys.space:SetKey(state)
	elseif(key=="LSHIFT" or key=="RSHIFT")then
		self.keys.shift:SetKey(state)
	elseif(key=="LCTRL" or key=="RCTRL")then
		self.keys.ctrl:SetKey(state)		
	elseif(key=="PAGEUP")then
		self.keys.pgup:SetKey(state)
	elseif(key=="PAGEDOWN")then
		self.keys.pgdn:SetKey(state)		
	elseif(key=="HOME")then
		self.keys.home:SetKey(state)
	elseif(key=="END")then
		self.keys.endkey:SetKey(state)		
	elseif(key=="TAB")then
		self.keys.tab:SetKey(state)
	elseif(key=="ESCAPE")then
		self.keys.esc:SetKey(state)		
	end
	
	if(self.environment_world~=nil and self.environment_world.player~=nil) then
		if(self.environment_world.player.dead==true) then
			self.environment_world.player=nil
		end
	end
	if(self.environment_world~=nil and self.environment_world.player~=nil) then	
		if(state==true)then			
			for k,v in ipairs(self.environment_world.player.combat.spellbuttons) do			
				if(v.key==key and v.shift==self.keys.shift.current and v.ctrl==self.keys.ctrl.current)then
					--print("Game.lua " .. tostring(v.spell))
					--print("Game.lua " .. v.spell.spellname)
					self.environment_world.player.combat:tryQueueSpell(v.spell)					
				end
			end
		end
	end
end

function TRAININGGROUNDS_Game:OnMouseChange(button,state)
	--print("TRAININGGROUNDS_Game:OnMouseChange "..button)
	if(self.mouse_input_enabled==true)then	
		--print("TRAININGGROUNDS_Game:OnMouseChange "..button)
		if(button=="LeftButton") then		
			self.keys.mouseL:SetKey(state)
		end
		if(button=="MiddleButton") then		
			self.keys.mouseM:SetKey(state)
		end				
		if(button=="RightButton") then		
			self.keys.mouseR:SetKey(state)
		end		
	end
end

function TRAININGGROUNDS_Game:OnMouseWheel(delta)
	--print("TRAININGGROUNDS_Game:OnMouseWheel "..delta)
	if(self.environment_world~=nil and self.environment_world.camera~=nil) then
		self.environment_world.camera:AdjustZoomViaScrollWheel(delta)
	end
end

function TRAININGGROUNDS_Game:OnEvent(event,...)
	if(event=="UI_SCALE_CHANGED")then
		--print("TRAININGGROUNDS_Game:OnEvent uiscalechanged")
		self:SetGameWindowScale()
	end
end



--TRAININGGROUNDS_Game.TEMP_MOVESPEED=105
TRAININGGROUNDS_Game.TEMP_MOVESPEED=105*4
TRAININGGROUNDS_Game.TEMP_DISTANCE=10	--used for model camera distance
	
	
	
function TRAININGGROUNDS_Game:step(elapsed)
	self.realtime=GetTime()*1000
	self.currenttime=self.currenttime+elapsed*1000
	--print(self.currenttimer)
	
	for i=1,#self.environments,1 do
		self.environments[i]:step(elapsed)
	end
	-- if(self.debugwindow~=nil)then
		-- for i=1,#self.debugwindow.environments,1 do
			-- self.debugwindow.environments[i]:step(elapsed)
		-- end	
	-- end
	
	
	-- clear pressed/released inputs
	self.keys:step()

	
	
	if(self.transitionfunction~=nil)then
		self.transitionprogress=self.transitionprogress+elapsed
		local a=math.sin(self.transitionprogress*math.pi)*1
		if(a>1)then a=1 end
		self.transitionfader.texture:SetTexture(0,0,0,a)
		if(self.transitionprogress>0.5)then
			if(self.transitiondone==false)then
				for i,v in ipairs(self.environments) do
					v:ClearAllObjects()
				end				
				self.debugwindow.frame:Hide()
				self.transitionfunction()
				self.transitiondone=true
				self.transitionfader.texture:SetTexture(0,0,0,1.0)
			end
		end		
		if(self.transitionprogress>1.0)then
			self.transitionprogress=1
			self.transitionfunction=nil
			self.transitionfader.texture:SetTexture(0,0,0,0)
		end
	end
end



function TRAININGGROUNDS_Game:draw()
	--self.player:draw()
end


function TRAININGGROUNDS_Game:CleanupGameplayEnvironmentsAfterReturningToMenu()
	--it turns out this is already handled by the transitionprogress check in step event
	--TODO: does cleanup actually happen, or do we only run ClearAllObjects

	-- self.environment_ui:Cleanup()
	-- self.environment_ui.frame:Hide()
	-- self.environment_menu:Cleanup()
	-- self.environment_menu.frame:Hide()	
	-- self.environment_world:Cleanup()
	-- self.environment_world.frame:Hide()		
	self:ShowFirstMenuAfterWarning()	
end

-- currently not used -- use createSpellTextbox instead
function TRAININGGROUNDS_Game:createSpellTextboxFromSpell(spell)
	local text=spell.spellname
	
	if(spell.usesgcd==false)then
		text="|cffffffff("..text..")"
	else
		--text="|cff00ff00"..text.."!"
		text="|cff00ff00"..text..""
		
		-- -- Hit Combo
		-- if(spell~=spell.combat.previousgcdspell)then
			-- text="|cff00ff00"..text.."!"
		-- else
			-- text="|cffffff00"..text
		-- end
		
		-- -- T19 4pc		
		-- if(spell~=spell.combat.previousgcdspell
		-- and spell~=spell.combat.previous2gcdspell
		-- and spell.combat.previousgcdspell~=spell.combat.previous2gcdspell
		-- and spell~=nil and spell.combat.previousgcdspell~=nil and spell.combat.previous2gcdspell~=nil)then		
			-- text="|cff00ff00"..text.."!"
		-- else
			-- text="|cffffff00"..text
		-- end
	end
	
	--text="|cff00ff00"..text.."!"
	--text="|cffffffff("..text..")"
	--text="|cffffff00"..text
	--text="|cffff0000nothing x2..."
	self:createSpellTextbox(text)
end

function TRAININGGROUNDS_Game:createSpellTextbox(text)
	--print("createSpellTextbox ",self.environment_ui)
	local textbox=TRAININGGROUNDS_TemporaryTextField.new()
	textbox:setup(self.environment_ui)
	textbox.frame.fontstring:SetJustifyH("LEFT")
	textbox.frame.fontstring:SetText(text)
	if(_TRAININGGROUNDS.debug_reversespellboxes==true)then
		textbox.x=25
		textbox.y=25
		textbox.xspeed=0
		textbox.yspeed=-20	
	else
		textbox.x=25
		textbox.y=-300
		textbox.xspeed=0
		textbox.yspeed=20	
	end
	
	textbox.frame:SetWidth(400)
	textbox:draw()
	textbox.frame:Show()
	

	--textbox.expirytime=self.environment_ui.currenttime+2000
	textbox.expirytime=self.environment_ui.currenttime+4000
	--print("Textbox expirytime: ",textbox.expirytime)
end

function TRAININGGROUNDS_Game:createBuffTextbox(text)
	local textbox=TRAININGGROUNDS_TemporaryTextField.new()
	textbox:setup(self.environment_ui)
	textbox.frame.fontstring:SetJustifyH("LEFT")
	textbox.frame.fontstring:SetText(text)
	if(_TRAININGGROUNDS.debug_reversespellboxes==true)then
		textbox.x=25
		textbox.y=25+10
		textbox.xspeed=0
		textbox.yspeed=-20	
	else
		textbox.x=25
		--textbox.y=-300+10		
		textbox.y=-300 -- don't need to at 10 at the moment because of spell impacttime
		textbox.xspeed=0
		textbox.yspeed=20
	end
	textbox.frame:SetWidth(400)
	textbox:draw()
	textbox.frame:Show()
	
		
	--textbox.expirytime=self.environment_ui.currenttime+2000
	textbox.expirytime=self.environment_ui.currenttime+4000+500
end

function TRAININGGROUNDS_Game:CreateComboTextbox(combocount)
	local textbox=TRAININGGROUNDS_TemporaryTextField.new()
	textbox:setup(self.environment_ui)
	textbox.frame.fontstring:SetJustifyH("LEFT")
	textbox.frame.fontstring:SetText(combocount.."-hit combo")
	textbox.x=350
	textbox.y=-75
	textbox:draw()
	textbox.frame:Show()
	
	textbox.xspeed=40
	textbox.yspeed=250
	textbox.xaccel=0
	textbox.yaccel=-400
	textbox.expirytime=self.environment_ui.currenttime+2000
end

function TRAININGGROUNDS_Game:ConvertLocalTimeToRealTime(localtime, environment)
	--TODO: realtime and localtime may drift over time
	-- if localtime is 1:1 with realtime, check for drift occasionally
		
	local offset=environment.currenttime-self.realtime
	--TODO: apply environment speedup/slowdown
	return localtime-offset
end

function TRAININGGROUNDS_Game:Cleanup()	
	if(self.debugwindow~=nil)then
		self.debugwindow:Cleanup()
	end
	self.transitionfader:Cleanup()
	self.border:Cleanup()
	--TODO:
	self.closebutton:RegisterEvent("OnClick",nil)
	TRAININGGROUNDS_ReusableButtonFrames.removeFrame(self.closebutton)
	

	for i,v in ipairs(self.environments) do
		v:Cleanup()
	end
	
	self.frame:Hide()
end