TRAININGGROUNDS_TypewriterMessage = TRAININGGROUNDS_inheritsFrom(TRAININGGROUNDS_TemporaryTextField)

-- typewriter messages are very awkwardly implemented.
-- a sizable amount of their features are handled in ScenarioController.

function TRAININGGROUNDS_TypewriterMessage.QuickSetup(scenario,message,expirytime,environment)
	if(environment==nil)then
		environment=_TRAININGGROUNDS.game.environment_ui
	end
		
	local tm=TRAININGGROUNDS_TypewriterMessage.new()
	tm:setup(environment)
	tm.scenario=scenario
	tm.frame:SetWidth(600);tm.frame:SetHeight(100)
	--tm.frame.fontstring:SetFont("Fonts\\FRIZQT__.TTF",24,"OUTLINE")
	tm.frame.fontstring:SetFont("Fonts\\FRIZQT__.TTF",20,"OUTLINE")
	tm:SetMessageAndCenterInUI(message)
	tm.y=-100
	tm.targety=-100
	tm.postexpirytime=expirytime
	
	-- if this is set to true, text will advance regardless of the status of any other text in the scenario
	-- 	and the scenario won't wait for this text to finish before changing state
	-- TODO: possibly allow for more than one typewriter manager per scenario
	
	tm.dontlock=false
	-- activation is handled by the owner scenario
	tm.activated=false

	tinsert(scenario.queuedtypewriters,tm)
	
	scenario.mostrecenttypewriter=tm
	
	return tm
end

TRAININGGROUNDS_TypewriterMessage.LETTERSPEED=32
TRAININGGROUNDS_TypewriterMessage.SPACESPEED=16

function TRAININGGROUNDS_TypewriterMessage:SetCustomInfo()
	TRAININGGROUNDS_TemporaryTextField.SetCustomInfo(self)
	
	self.targetmessage=""
	self.letterindex=0
	self.nextletterindex=1
	--self.nextlettertick=0
	self.nextlettertick=self.environment.currenttime
	
	-- this is actually a divisor, not a multiplier
	-- higher divisor means higher speed.
	-- divisor of 0 pauses message.
	-- divisor of nil reveals full message instantly.
	self.speedmultiplier=1
	
	-- this expirytime will be set only after the message is done typing.
	self.postexpirytime=nil
end

function TRAININGGROUNDS_TypewriterMessage:SetAppearance()
	
	--self.frame:Show()	
	self.frame.tag="TemporaryTextField"
	self.frame.fontstring:Show()	
	--print("TemporaryTextField:SetAppearance "..self.text)	
	self.alignment="TOPLEFT"
end

function TRAININGGROUNDS_TypewriterMessage:SetFrameSize()
	-- don't call super; text doesn't depend on camera zoom
	-- TODO: resize text based on camera zoom, maybe
end

function TRAININGGROUNDS_TypewriterMessage:SetMessageAndCenterInUI(message)
	-- this function sets x-coordinate, but not y-coordinate
	
	self.targetmessage=message
	self:CenterInUI()
	
end

function TRAININGGROUNDS_TypewriterMessage:CenterInUI()
	-- this currently expects letterindex == 0

	self.frame.fontstring:SetText(self.targetmessage)
	self.frame.fontstring:SetJustifyH("LEFT")
	local width=self.frame.fontstring:GetStringWidth()
	self.textwidth=width
	self.textheight=self.frame.fontstring:GetStringHeight()
	-- if message is (e.g.) empty, set height manually
	if(self.textheight==0)then
		--TODO: const
		self.textheight=20
	end
	self.frame.fontstring:SetText("")
	local uiwidth=TRAININGGROUNDS_Game.SCREENWIDTH
	-- alignment is TOPLEFT
	self.x=uiwidth/2-width/2
	--self.frame.texture:Show()
	--self.frame.framelevel=500
	self.frame.framelevel=120  --!!!
	self.frame:SetFrameLevel(self.frame.framelevel)
	--print(self.x)--!!!
	--self.x=0
	

end

function TRAININGGROUNDS_TypewriterMessage:step(elapsed)
	TRAININGGROUNDS_TemporaryTextField.step(self,elapsed)
	
	local currenttime=self.environment.currenttime
	
	-- --!!!
	-- if(_TRAININGGROUNDS.game.keys.ctrl.current==true)then	
		-- print("TypewriterMessage:step ",currenttime,self.targetmessage,self.expirytime)
	-- end
	
	
	if(self.speedmultiplier==nil)then
		
	end
	
	if(self.speedmultiplier~=0 and (self.dontlock==true or self.activated==true))then		
		local lettercount=0
		while(currenttime>=self.nextlettertick and lettercount<100)do
			lettercount=lettercount+1
			--if(self.letterindex<#self.targetmessage)then
			if(self.letterindex<#self.targetmessage)then
			
		
				self.letterindex=self.nextletterindex			
				-- colors won't appear unless we have a close tag
				-- but there's no penalty for having an extra close tag, so just add a close tag to the current message
				local message=string.sub(self.targetmessage,1,self.letterindex).."|r"
				self.frame.fontstring:SetText(message)
				
				
				if(self.letterindex<#self.targetmessage)then
					self.nextletterindex=self.letterindex+1
					
					local nextletter=string.sub(self.targetmessage,self.nextletterindex,self.nextletterindex)
					local speed=0
					if(nextletter=="|")then
						-- escape sequence (only color escape sequences are supported at the moment)					
						self.nextletterindex=self.nextletterindex+10
						nextletter=string.sub(self.targetmessage,self.nextletterindex,self.nextletterindex)
					end
					--print(nextletter)
					if(nextletter==" ")then
						speed=TRAININGGROUNDS_TypewriterMessage.SPACESPEED
					else
						speed=TRAININGGROUNDS_TypewriterMessage.LETTERSPEED
					end
					if(self.speedmultiplier==nil)then
						-- speedmultiplier==nil check is handled further up the page
					elseif(self.speedmultiplier==0)then
						-- the pause-message check is handled further up the page
						self.nextlettertick=currenttime						
					else
						
						speed=speed/self.speedmultiplier
						--self.nextlettertick=currenttime+speed
						self.nextlettertick=self.nextlettertick+speed
					end
				end
			else
				-- break while loop without going through all 100 iterations...
				self.nextlettertick=currenttime+1
			end
		end
	else
		--else time isn't ticking yet
		self.nextlettertick=self.nextlettertick+elapsed*1000
	end
	
	-- --!!!
	-- if(self.targetmessage=="")then
		-- print(self.letterindex,#self.targetmessage,self.expirytime,self.postexpirytime)
	-- end
	
	if(self.letterindex>=#self.targetmessage)then
		if(self.expirytime==nil and self.postexpirytime~=nil)then
			self.expirytime=currenttime+self.postexpirytime
		end
	end
		
	--if(self.expirytime~=nil and self.expirytime<currenttime+1000)then
		--self.yspeed=30
	--end
end










